#
# Copyright 2017-2023 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Get and set Renku repository or global options.

Description
~~~~~~~~~~~

Save useful Renku configuration directives in a simple human-readable ini file.
The values are used by other renku commands and by interactive sessions.

Check the `Available configuration values`_ section to get more information
on the available values.


Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.config:config
   :prog: renku config
   :nested: full

Examples
~~~~~~~~

.. cheatsheet::
   :group: Config
   :command: $ renku config show [<key>]
   :description: Show current configuration.
   :target: rp

You can set various Renku configuration options, for example the image registry
URL, with a command like:

.. code-block:: console

    $ renku config set interactive.default_url "/tree"

By default, configuration is stored locally in the project's directory. Use
``--global`` option to store configuration for all projects in your home
directory.

.. cheatsheet::
   :group: Config
   :command: $ renku config set <key> <value>
   :description: Set entry <key> to <value> in renku config.
   :target: rp

To remove a specific key from configuration use:

.. code-block:: console

    $ renku config remove interactive.default_url

By default, only local configuration is searched for removal. Use ``--global``
option to remove a global configuration value.

.. cheatsheet::
   :group: Config
   :command: $ renku config remove <key>
   :description: Unset entry <key> renku config.
   :target: rp

Available configuration values
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The following values are used by renku-python and available for
the ``renku config`` command:

+--------------------------------+-------------------------------------+-----------+
| Name                           | Description                         | Default   |
+================================+=====================================+===========+
| ``data_directory``             | Path to the data directory          | ``data/`` |
|                                | (read-only after project creation)  |           |
+--------------------------------+-------------------------------------+-----------+
| ``dataverse.access_token``     | Access token for Dataverse API      | ``None``  |
+--------------------------------+-------------------------------------+-----------+
| ``dataverse.server_url``       | URL for the Dataverse API server    | ``None``  |
|                                | to use                              |           |
+--------------------------------+-------------------------------------+-----------+
| ``default_dataset_add_action`` | Default action when adding files to | ``None``  |
|                                | datasets. Can be either ``copy`` or |           |
|                                | ``move``.                           |           |
+--------------------------------+-------------------------------------+-----------+
| ``lfs_threshold``              | Threshold file size below which     | ``100kb`` |
|                                | files are not added to git LFS      |           |
+--------------------------------+-------------------------------------+-----------+
| ``show_lfs_message``           | Whether to show messages about      | ``true``  |
|                                | files being added to git LFS or not |           |
+--------------------------------+-------------------------------------+-----------+
| ``show_login_warning``         | Whether to warn when logging in to  | ``true``  |
|                                | Renku inside a project that causes  |           |
|                                | project's git remote to be changed. |           |
+--------------------------------+-------------------------------------+-----------+
| ``ssh_supported``              | Whether the image in this project   | ``false`` |
|                                | contains an SSH server on port 22   |           |
|                                | for SSH connections. Used to        |           |
|                                | override the corresponding project  |           |
|                                | template setting.                   |           |
+--------------------------------+-------------------------------------+-----------+
| ``zenodo.access_token``        | Access token for Zenodo API         | ``None``  |
+--------------------------------+-------------------------------------+-----------+

See the section on `renku.ini <https://renku.readthedocs.io/en/latest/reference/templates.html#renku>`_
for more configuration values.


"""
import click

import renku.ui.cli.utils.color as color
from renku.ui.cli.utils.click import MutuallyExclusiveOption


@click.group()
def config():
    """Interact with renku configuration."""
    pass


@config.command()
@click.argument("key", required=False, default=None)
@click.option(
    "--local",
    "local_only",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    help="Read from local configuration only.",
    mutually_exclusive=[("global_only", "--global"), ("default_only", "--default")],
)
@click.option(
    "--global",
    "global_only",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    help="Read from global configuration only.",
    mutually_exclusive=[("local_only", "--local"), ("default_only", "--default")],
)
@click.option(
    "--default",
    "default_only",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    help="Show default values if applicable.",
    mutually_exclusive=[("local_only", "--local"), ("global_only", "--global")],
)
def show(key, local_only, global_only, default_only):
    """Show current configuration.

    KEY is of the form <group>.<entry>, e.g. 'interactive.default_url'.
    """
    from renku.command.config import read_config
    from renku.domain_model.enums import ConfigFilter

    config_filter = ConfigFilter.ALL

    if local_only:
        config_filter = ConfigFilter.LOCAL_ONLY
    elif global_only:
        config_filter = ConfigFilter.GLOBAL_ONLY
    elif default_only:
        config_filter = ConfigFilter.DEFAULT_ONLY

    value = read_config().build().execute(key, config_filter=config_filter)
    click.secho(value.output)


@config.command("set")
@click.argument("key")
@click.argument("value")
@click.option("--global", "global_only", is_flag=True, help="Store to global configuration only.")
def set_(key, value, global_only):
    """Set a configuration value.

    KEY is of the form <group>.<entry>, e.g. 'interactive.default_url'.
    """
    from renku.command.config import update_config

    update_config().build().execute(key, value=value, global_only=global_only)
    click.secho("OK", fg=color.GREEN)


@config.command()
@click.argument("key")
@click.option("--global", "global_only", is_flag=True, help="Remove from global configuration only.")
def remove(key, global_only):
    """Remove a configuration value.

    KEY is of the form <group>.<entry>, e.g. 'interactive.default_url'.
    """
    from renku.command.config import update_config

    update_config().build().execute(key, remove=True, global_only=global_only)
    click.secho("OK", fg=color.GREEN)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""WorkflowFile view models."""

from __future__ import annotations

import dataclasses
from datetime import datetime
from typing import Any, List, Optional

from renku.core.workflow.model.workflow_file import HiddenParameter, Input, Output, Parameter, Step, WorkflowFile


@dataclasses.dataclass
class WorkflowFileViewModel:
    """View model for ``WorkflowFile``."""

    description: Optional[str]
    keywords: List[str]
    name: str
    path: str
    steps: List[StepViewModel]

    @classmethod
    def from_workflow_file(cls, workflow_file: WorkflowFile) -> WorkflowFileViewModel:
        """Create an instance from a ``WorkflowFile``."""
        return cls(
            description=workflow_file.description,
            keywords=workflow_file.keywords.copy(),
            name=workflow_file.name,
            path=workflow_file.path,
            steps=[StepViewModel.from_step(s) for s in workflow_file.steps],
        )


@dataclasses.dataclass
class StepViewModel:
    """View models for ``Step``."""

    command: str
    full_command: str
    date_created: datetime
    description: Optional[str]
    inputs: List[InputViewModel]
    keywords: List[str]
    name: str
    outputs: List[OutputViewModel]
    parameters: List[ParameterViewModel]
    success_codes: List[int]

    @classmethod
    def from_step(cls, step: Step) -> StepViewModel:
        """Create an instance from a ``Step``."""
        return cls(
            command=step.command,
            full_command=step.original_command,
            date_created=step.date_created,
            description=step.description,
            inputs=[InputViewModel.from_input(i) for i in step.inputs],
            keywords=step.keywords.copy(),
            name=step.name,
            outputs=[OutputViewModel.from_output(o) for o in step.outputs],
            parameters=[
                ParameterViewModel.from_parameter(p) for p in step.parameters if not isinstance(p, HiddenParameter)
            ],
            success_codes=step.success_codes.copy(),
        )


@dataclasses.dataclass
class ParameterViewModel:
    """View model for ``Parameter``."""

    description: Optional[str]
    implicit: bool
    name: Optional[str]
    position: Optional[int]
    prefix: Optional[str]
    value: Any

    @classmethod
    def from_parameter(cls, parameter: Parameter) -> ParameterViewModel:
        """Create an instance from a ``Parameter``."""
        return cls(
            description=parameter.description,
            implicit=parameter.implicit,
            name=parameter.name,
            position=parameter.position,
            prefix=parameter.prefix,
            value=parameter.value,
        )


@dataclasses.dataclass
class InputViewModel:
    """View model for ``Input``."""

    description: Optional[str]
    implicit: bool
    mapped_to: Optional[str]
    name: Optional[str]
    path: str
    position: Optional[int]
    prefix: Optional[str]

    @classmethod
    def from_input(cls, input: Input) -> InputViewModel:
        """Create an instance from an ``Input``."""
        return cls(
            description=input.description,
            implicit=input.implicit,
            mapped_to=input.mapped_to,
            name=input.name,
            path=input.path,
            position=input.position,
            prefix=input.prefix,
        )


@dataclasses.dataclass
class OutputViewModel:
    """View model for ``Output``."""

    description: Optional[str]
    implicit: bool
    mapped_to: Optional[str]
    name: Optional[str]
    path: str
    persist: bool
    position: Optional[int]
    prefix: Optional[str]

    @classmethod
    def from_output(cls, output: Output) -> OutputViewModel:
        """Create an instance from an ``Output``."""
        return cls(
            description=output.description,
            implicit=output.implicit,
            mapped_to=output.mapped_to,
            name=output.name,
            path=output.path,
            persist=output.persist,
            position=output.position,
            prefix=output.prefix,
        )



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Datasets Provenance."""

from datetime import datetime
from typing import TYPE_CHECKING, List, Literal, Optional, Union, overload
from uuid import UUID

from renku.command.command_builder.command import inject
from renku.core import errors
from renku.core.interface.dataset_gateway import IDatasetGateway
from renku.core.util import communication

if TYPE_CHECKING:
    from renku.domain_model.dataset import Dataset, DatasetTag
    from renku.domain_model.provenance.agent import Person


class DatasetsProvenance:
    """A set of datasets."""

    dataset_gateway = inject.attr(IDatasetGateway)

    @property
    def datasets(self) -> List["Dataset"]:
        """Return an iterator of datasets."""
        return self.dataset_gateway.get_all_active_datasets()

    def get_by_id(self, id: str, immutable: bool = False) -> Optional["Dataset"]:
        """Return a dataset by its id."""
        from renku.domain_model.dataset import Dataset

        try:
            dataset = self.dataset_gateway.get_by_id(id)
        except errors.ObjectNotFoundError:
            pass
        else:
            assert isinstance(dataset, Dataset)
            if not dataset.immutable or immutable:
                return dataset

            return dataset.copy()
        return None

    @overload
    def get_by_name(self, name: str, *, immutable: bool = False, strict: Literal[False] = False) -> Optional["Dataset"]:
        ...

    @overload
    def get_by_name(self, name: str, *, immutable: bool = False, strict: Literal[True]) -> "Dataset":
        ...

    def get_by_name(
        self, name: str, immutable: bool = False, strict: bool = False
    ) -> Union[Optional["Dataset"], "Dataset"]:
        """Return a dataset by its name.

        Args:
            name(str): Name of the dataset
            immutable(bool): Whether the dataset will be used as an immutable instance or will be modified (Default
                value = False).
            strict(bool): Whether to raise an exception if the dataset doesn't exist or not (Default value = False)

        Returns:
            Optional[Dataset]: Dataset with the specified name if exists.
        """
        dataset = self.dataset_gateway.get_by_name(name)
        if not dataset:
            if strict:
                raise errors.DatasetNotFound(name=name)

            return None
        if not dataset.immutable or immutable:
            return dataset

        return dataset.copy()

    def get_provenance_tails(self):
        """Return the provenance for all datasets."""
        return self.dataset_gateway.get_provenance_tails()

    def get_previous_version(self, dataset: "Dataset") -> Optional["Dataset"]:
        """Return the previous version of a dataset if any."""
        if dataset.is_derivation():
            return self.get_by_id(dataset.derived_from.url_id)  # type: ignore

        return None

    def add_or_update(self, dataset: "Dataset", date: Optional[datetime] = None, creator: Optional["Person"] = None):
        """Add/update a dataset according to its new content.

        NOTE: This functions always mutates the dataset.
        """
        from renku.domain_model.dataset import Dataset

        assert isinstance(dataset, Dataset)

        # NOTE: Dataset's name never changes, so, we use it to detect if a dataset should be mutated.
        current_dataset = self.get_by_name(dataset.name)

        if current_dataset:
            assert (
                not current_dataset.is_removed()
            ), f"Adding/Updating a removed dataset '{dataset.name}:{dataset.identifier}'"

            dataset.update_files_from(current_dataset, date=date)

            # NOTE: Always mutate a dataset to make sure an old identifier is not reused
            dataset.derive_from(current_dataset, creator=creator)
        else:
            assert (
                dataset.derived_from is None
            ), f"Parent dataset {dataset.derived_from} not found for '{dataset.name}:{dataset.identifier}'"

        self.dataset_gateway.add_or_remove(dataset)

    def remove(self, dataset, date: Optional[datetime] = None, creator: Optional["Person"] = None):
        """Remove a dataset."""
        from renku.domain_model.dataset import Dataset

        assert isinstance(dataset, Dataset)

        # NOTE: Dataset's name never changes, so, we use it to detect if a dataset should be mutated.
        current_dataset = self.dataset_gateway.get_by_name(dataset.name)

        if current_dataset:
            assert not current_dataset.is_removed(), f"Removing a removed dataset '{dataset.name}:{dataset.identifier}'"

            # NOTE: We always assign a new identifier to make sure an old identifier is not reused
            dataset.derive_from(current_dataset, creator=creator)
        else:
            assert (
                dataset.derived_from is None
            ), f"Parent dataset {dataset.derived_from} not found for '{dataset.name}:{dataset.identifier}'"

        dataset.remove(date)
        self.dataset_gateway.add_or_remove(dataset)

    def update_during_migration(
        self,
        dataset: "Dataset",
        commit_sha: str,
        date: Optional[datetime] = None,
        tags: Optional[List["DatasetTag"]] = None,
        remove=False,
        replace=False,
        preserve_identifiers: bool = False,
    ):
        """Add, update, remove, or replace a dataset in migration."""
        from renku.domain_model.dataset import Dataset

        assert isinstance(dataset, Dataset)
        assert not (remove and replace), "Cannot remove and replace"

        def update_dataset(existing, new) -> "Dataset":
            """Update existing dataset with the new dataset metadata."""
            existing.update_metadata_from(new, exclude=["date_created", "derived_from", "same_as"])
            existing.dataset_files = new.dataset_files
            return existing

        # NOTE: Dataset's name never changes, so, we use it to detect if a dataset should be mutated.
        current_dataset = self.get_by_name(dataset.name, immutable=True)

        new_identifier = self._create_dataset_identifier(commit_sha, dataset.identifier)
        dataset_with_same_id = self.get_by_id(dataset.id, immutable=True)

        if dataset_with_same_id and preserve_identifiers:
            dataset.update_files_from(dataset_with_same_id, date=date)

            dataset = update_dataset(existing=dataset_with_same_id, new=dataset)
        elif current_dataset:
            dataset.update_files_from(current_dataset, date=date)

            if replace:
                # NOTE: Copy metadata to the current dataset
                dataset = update_dataset(existing=current_dataset, new=dataset)
            else:
                identifier = new_identifier if dataset_with_same_id else dataset.identifier
                date_created = date if dataset_with_same_id else dataset.date_created
                dataset.derive_from(current_dataset, creator=None, identifier=identifier, date_created=date_created)
        else:
            if remove:
                communication.warn(f"Deleting non-existing dataset '{dataset.name}'")

            if dataset.derived_from:
                communication.warn(
                    f"Parent dataset {dataset.derived_from} not found for '{dataset.name}:{dataset.identifier}'"
                )
                dataset.derived_from = None

            # NOTE: This happens in migrations of broken projects
            if dataset_with_same_id:
                dataset.replace_identifier(new_identifier)

        if remove:
            dataset.remove()
        else:
            self._process_dataset_tags(dataset, tags)

        self.dataset_gateway.add_or_remove(dataset)

    @staticmethod
    def _create_dataset_identifier(commit_sha: str, identifier: str) -> str:
        uuid = f"{commit_sha[:20]}{identifier[-12:]}"
        return UUID(uuid).hex

    def get_all_tags(self, dataset: "Dataset") -> List["DatasetTag"]:
        """Return the list of all tags for a dataset."""
        return self.dataset_gateway.get_all_tags(dataset)

    def add_tag(self, dataset: "Dataset", tag: "DatasetTag"):
        """Add a tag from a dataset."""
        self.dataset_gateway.add_tag(dataset, tag)

    def remove_tag(self, dataset: "Dataset", tag: "DatasetTag"):
        """Remove a tag from a dataset."""
        self.dataset_gateway.remove_tag(dataset, tag)

    def _process_dataset_tags(self, dataset: "Dataset", tags: Optional[List["DatasetTag"]]):
        from renku.domain_model.dataset import DatasetTag, Url

        if not tags:
            return

        current_tag_names = [t.name for t in self.get_all_tags(dataset)]
        for tag in tags:
            if tag.name in current_tag_names:
                continue
            tag = DatasetTag(
                dataset_id=Url(url_id=dataset.id),
                date_created=tag.date_created,
                description=tag.description,
                name=tag.name,
            )
            self.add_tag(dataset, tag)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility for working with HTTP session.

This module provides some wrapper functions around ``requests`` library. It sets a timeout and converts exception types
whenever needed. Use this module instead of ``requests``.
"""

import os
import tempfile
import urllib
from contextlib import contextmanager
from pathlib import Path
from typing import Union

import patoolib
import requests
from requests.adapters import HTTPAdapter, TimeoutSauce  # type: ignore
from urllib3.util.retry import Retry

from renku.core import errors

_RENKU_REQUESTS_TIMEOUT_SECONDS = float(os.getenv("RENKU_REQUESTS_TIMEOUT_SECONDS", 1200))


class _CustomTimeout(TimeoutSauce):
    """CustomTimeout for all HTTP requests."""

    def __init__(self, *args, **kwargs):
        """Construct CustomTimeout."""
        if kwargs["connect"] is None:
            kwargs["connect"] = _RENKU_REQUESTS_TIMEOUT_SECONDS

        if kwargs["read"] is None:
            kwargs["read"] = _RENKU_REQUESTS_TIMEOUT_SECONDS

        super().__init__(*args, **kwargs)


requests.adapters.TimeoutSauce = _CustomTimeout  # type: ignore


def delete(url, headers=None):
    """Send a DELETE request."""
    return _request("delete", url=url, headers=headers)


def get(url, headers=None, params=None):
    """Send a GET request."""
    return _request("get", url=url, headers=headers, params=params)


def head(url, *, allow_redirects=False, headers=None):
    """Send a HEAD request."""
    return _request("head", url=url, allow_redirects=allow_redirects, headers=headers)


def post(url, *, data=None, files=None, headers=None, json=None, params=None):
    """Send a POST request."""
    return _request("post", url=url, data=data, files=files, headers=headers, json=json, params=params)


def put(url, *, data=None, files=None, headers=None, params=None):
    """Send a PUT request."""
    return _request("put", url=url, data=data, files=files, headers=headers, params=params)


def _request(verb: str, url: str, *, allow_redirects=True, data=None, files=None, headers=None, json=None, params=None):
    try:
        with _retry() as session:
            return getattr(session, verb)(
                url=url,
                allow_redirects=allow_redirects,
                data=data,
                files=files,
                headers=headers,
                json=json,
                params=params,
            )
    except (ConnectionError, requests.RequestException, urllib.error.HTTPError) as e:
        raise errors.RequestError(f"{verb.upper()} request failed for {url}") from e


def get_redirect_url(url) -> str:
    """Return redirect URL if any; otherwise, return the original URL."""
    try:
        return head(url, allow_redirects=True).url
    except errors.RequestError:
        # NOTE: HEAD request failed, try with original url
        return url


def check_response(response):
    """Check for expected response status code."""
    if response.status_code in [200, 201, 202]:
        return
    elif response.status_code == 401:
        raise errors.AuthenticationError("Access unauthorized - update access token.")
    else:
        content = response.content.decode("utf-8") if response.content else ""
        message = f"Request failed with code {response.status_code}: {content}"
        raise errors.RequestError(message)


def download_file(base_directory: Union[Path, str], url: str, filename, extract, chunk_size=16384):
    """Download a URL to a given location."""
    from renku.core.util import communication

    def extract_dataset(filepath):
        """Extract downloaded file."""
        try:
            tmp = tempfile.mkdtemp()
            patoolib.extract_archive(str(filepath), outdir=tmp, verbosity=-1)
        except patoolib.util.PatoolError:
            return filepath.parent, [filepath]
        else:
            filepath.unlink()
            return Path(tmp), [p for p in Path(tmp).rglob("*")]

    tmp_root = Path(base_directory)
    tmp_root.mkdir(parents=True, exist_ok=True)
    tmp = tempfile.mkdtemp(dir=tmp_root)

    try:
        with requests.get(url, stream=True, allow_redirects=True) as response:
            response.raise_for_status()

            if not filename:
                filename = get_filename_from_headers(response)

            if not filename:
                u = urllib.parse.urlparse(url)
                filename = Path(u.path).name
                if not filename:
                    raise errors.ParameterError(f"URL Cannot find a file to download from {url}")

            download_to = Path(tmp) / filename
            with open(str(download_to), "wb") as file_:
                total_size = int(response.headers.get("content-length", 0))

                communication.start_progress(name=filename, total=total_size)
                try:
                    for chunk in response.iter_content(chunk_size=chunk_size):
                        if chunk:  # ignore keep-alive chunks
                            file_.write(chunk)
                            communication.update_progress(name=filename, amount=len(chunk))
                finally:
                    communication.finalize_progress(name=filename)

    except (requests.exceptions.HTTPError, urllib.error.HTTPError) as e:  # pragma nocover
        raise errors.RequestError(f"Cannot download from {url}") from e

    if extract:
        return extract_dataset(download_to)

    return download_to.parent, [download_to]


def get_filename_from_headers(response):
    """Extract filename from content-disposition headers if available."""
    content_disposition = response.headers.get("content-disposition", None)

    if not content_disposition:
        return None

    entries = content_disposition.split(";")
    name_entry = next((e.strip() for e in entries if e.strip().lower().startswith("filename*=")), None)

    if name_entry:
        name = name_entry.split("=", 1)[1].strip()
        encoding, _, name = name.split("'")
        return urllib.parse.unquote(name, encoding, errors="strict")

    name_entry = next((e.strip() for e in entries if e.strip().lower().startswith("filename=")), None)

    if not name_entry:
        return None

    filename = name_entry.split("=", 1)[1].strip()

    if filename.startswith('"'):
        filename = filename[1:-1]
    return filename


@contextmanager
def _retry(total_requests=5, backoff_factor=0.2, statuses=(500, 502, 503, 504, 429)):
    """Default HTTP session for requests."""
    _session = requests.Session()

    retries = Retry(total=total_requests, backoff_factor=backoff_factor, status_forcelist=list(statuses))

    _session.mount("http://", HTTPAdapter(max_retries=retries))
    _session.mount("https://", HTTPAdapter(max_retries=retries))

    try:
        yield _session
    except requests.RequestException as e:
        raise errors.RequestError("renku operation failed due to network connection failure") from e



#
# Copyright 2017-2023 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku RDF Graph API.

The ``RDFGraph`` class allows for the quick creation of a searchable graph object
based on the project's metadata.

To create the graph and query it:

.. code-block:: python

    from renku.ui.api import RDFGraph

    g = RDFGraph()
    # get a list of contributors to the project
    list(g.subjects(object=URIRef("http://schema.org/Person")))

For more information on querying the graph, see the `RDFLib
documentation <https://rdflib.readthedocs.io/en/stable/intro_to_graphs.html>`_.

"""

import json

from rdflib import Graph

from renku.command.graph import export_graph_command


class RDFGraph(Graph):
    """RDF Graph of the project's metadata."""

    def __init__(self, revision_or_range=None):
        """Instantiate the RDFGraph class.

        Args:
            revision_or_range(None): Revision or range to generate the graph from. Defaults to ``None``

        """
        super().__init__()
        self.revision_or_range = revision_or_range
        self._build()
        self.bind_(self)

    def _build(self):
        """Construct the RDF graph representing this Renku project."""
        import pyld

        data = json.dumps(
            pyld.jsonld.expand(
                export_graph_command().build().execute(revision_or_range=self.revision_or_range).output._graph
            )
        )
        self.parse(data=data, format="json-ld")

    @staticmethod
    def bind_(graph):
        """Bind all the usual namespaces."""
        graph.bind("prov", "http://www.w3.org/ns/prov#")
        graph.bind("oa", "http://www.w3.org/ns/oa#")
        graph.bind("schema", "http://schema.org/")
        graph.bind("renku", "https://swissdatasciencecenter.github.io/renku-ontology#")
        graph.bind("foaf", "http://xmlns.com/foaf/0.1/")



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Project business logic."""

from typing import Dict, List, Optional, Union, cast

from pydantic import validate_arguments

from renku.command.command_builder import inject
from renku.command.view_model.project import ProjectViewModel
from renku.core.interface.project_gateway import IProjectGateway
from renku.core.util.metadata import construct_creator
from renku.domain_model.constant import NO_VALUE, NoValueType
from renku.domain_model.project_context import project_context
from renku.domain_model.provenance.agent import Person


@inject.autoparams()
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def edit_project(
    description: Optional[Union[str, NoValueType]],
    creator: Union[Dict, str, NoValueType],
    keywords: Optional[Union[List[str], NoValueType]],
    custom_metadata: Optional[Union[Dict, List[Dict], NoValueType]],
    custom_metadata_source: Optional[Union[str, NoValueType]],
    project_gateway: IProjectGateway,
):
    """Edit dataset metadata.

    Args:
        description(Union[Optional[str], NoValueType]): New description.
        creator(Union[Dict, str, NoValueType]): New creators.
        keywords(Union[Optional[List[str]]): New keywords.
        custom_metadata(Union[Optional[Dict, List[Dict]]): Custom JSON-LD metadata.
        custom_metadata_source(Optional[str]): Custom metadata source.
        project_gateway(IProjectGateway): Injected project gateway.

    Returns:
        Tuple of fields that were updated and dictionary of warnings.
    """
    possible_updates = {
        "creator": creator,
        "description": description,
        "keywords": keywords,
        "custom_metadata": custom_metadata,
    }

    no_email_warnings: Optional[Union[Dict, str]] = None
    parsed_creator: Optional[Union[NoValueType, Person]] = NO_VALUE

    if creator is not NO_VALUE:
        parsed_creator, no_email_warnings = construct_creator(cast(Union[Dict, str], creator), ignore_email=True)

    updated = {k: v for k, v in possible_updates.items() if v is not NO_VALUE}

    if updated:
        project = project_gateway.get_project()
        project.update_metadata(
            creator=parsed_creator,
            description=description,
            keywords=keywords,
            custom_metadata=custom_metadata,
            custom_metadata_source=custom_metadata_source,
        )
        project_gateway.update_project(project)

    return updated, no_email_warnings


def show_project() -> ProjectViewModel:
    """Show project metadata.

    Returns:
        Project view model.
    """
    return ProjectViewModel.from_project(project_context.project)



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service project view."""
from flask import request

from renku.ui.service.config import SERVICE_PREFIX
from renku.ui.service.controllers.project_edit import ProjectEditCtrl
from renku.ui.service.controllers.project_lock_status import ProjectLockStatusCtrl
from renku.ui.service.controllers.project_show import ProjectShowCtrl
from renku.ui.service.views.api_versions import ALL_VERSIONS, VersionedBlueprint
from renku.ui.service.views.decorators import accepts_json, requires_cache, requires_identity
from renku.ui.service.views.error_handlers import handle_common_except, handle_project_write_errors

PROJECT_BLUEPRINT_TAG = "project"
project_blueprint = VersionedBlueprint(PROJECT_BLUEPRINT_TAG, __name__, url_prefix=SERVICE_PREFIX)


@project_blueprint.route("/project.show", methods=["POST"], provide_automatic_options=False, versions=ALL_VERSIONS)
@handle_common_except
@accepts_json
@requires_cache
@requires_identity
def show_project_view(user_data, cache):
    """
    Show project metadata view.

    ---
    post:
      description: Show project metadata.
      requestBody:
        content:
          application/json:
            schema: ProjectShowRequest
      responses:
        200:
          description: Metadata of the project.
          content:
            application/json:
              schema: ProjectShowResponseRPC
      tags:
        - project
    """
    return ProjectShowCtrl(cache, user_data, dict(request.json)).to_response()  # type: ignore


@project_blueprint.route("/project.edit", methods=["POST"], provide_automatic_options=False, versions=ALL_VERSIONS)
@handle_common_except
@handle_project_write_errors
@accepts_json
@requires_cache
@requires_identity
def edit_project_view(user_data, cache):
    """
    Edit project metadata view.

    Not passing a field leaves it unchanged.

    ---
    post:
      description: Edit project metadata.
      requestBody:
        content:
          application/json:
            schema: ProjectEditRequest
      responses:
        200:
          description: Status of the requested project edits.
          content:
            application/json:
              schema: ProjectEditResponseRPC
      tags:
        - project
    """
    return ProjectEditCtrl(cache, user_data, dict(request.json)).to_response()  # type: ignore


@project_blueprint.route(
    "/project.lock_status", methods=["GET"], provide_automatic_options=False, versions=ALL_VERSIONS
)
@handle_common_except
@requires_cache
@requires_identity
def get_project_lock_status(user_data, cache):
    """
    Check whether a project is locked for writing or not.

    ---
    get:
      description: Get project write-lock status.
      parameters:
        - in: query
          schema: ProjectLockStatusRequest
      responses:
        200:
          description: Status of the project write-lock.
          content:
            application/json:
              schema: ProjectLockStatusResponseRPC
      tags:
        - project
    """
    return ProjectLockStatusCtrl(cache, user_data, dict(request.args)).to_response()



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service parent serializers."""
import uuid
from datetime import datetime

import yagup
from marshmallow import Schema, fields, pre_load, validates

from renku.ui.service.errors import UserRepoUrlInvalidError
from renku.ui.service.serializers.rpc import JsonRPCResponse


class LocalRepositorySchema(Schema):
    """Schema for identifying a locally stored repository."""

    # In the long term, the id should be used only for internal operations
    project_id = fields.String(metadata={"description": "Reference to access the project in the local cache."})


class RemoteRepositoryBaseSchema(Schema):
    """Schema for tracking a remote repository."""

    git_url = fields.String(metadata={"description": "Remote git repository url."})

    @validates("git_url")
    def validate_git_url(self, value):
        """Validates git url."""
        if value:
            try:
                yagup.parse(value)
            except yagup.exceptions.InvalidURL as e:
                raise UserRepoUrlInvalidError(e, "Invalid `git_url`")

        return value


class RemoteRepositorySchema(RemoteRepositoryBaseSchema):
    """Schema for tracking a remote repository and branch."""

    branch = fields.String(load_default=None, metadata={"description": "Remote git branch (or tag or commit SHA)."})

    @pre_load
    def set_branch_from_ref(self, data, **kwargs):
        """Set `branch` field from `ref` if present."""
        if "ref" in data and not data.get("branch"):
            # Backward compatibility: branch and ref were both used. Let's keep branch as the exposed field
            # even if internally it gets converted to "ref" later.
            data["branch"] = data["ref"]
            del data["ref"]

        return data


class AsyncSchema(Schema):
    """Schema for adding a commit at the end of the operation."""

    is_delayed = fields.Boolean(metadata={"description": "Whether the job should be delayed or not."})


class MigrateSchema(Schema):
    """Schema for allowing preliminary repository migration."""

    migrate_project = fields.Boolean(
        default=False,
        load_default=False,
        metadata={"description": "Whether the project should be migrated before the other operations take place."},
    )


class ArchiveSchema(Schema):
    """Schema for unpacking archives."""

    unpack_archive = fields.Boolean(
        load_default=False,
        metadata={"description": "Whether to automatically extract archive content."},
    )


class MandatoryUserSchema(Schema):
    """Schema for adding user as requirement."""

    user_id = fields.String(
        required=True,
        metadata={"description": "Mandatory user id."},
    )


class CreationSchema(Schema):
    """Schema for creation date."""

    created_at = fields.DateTime(
        load_default=datetime.utcnow,
        metadata={"description": "Creation date."},
    )


class FileDetailsSchema(ArchiveSchema, CreationSchema):
    """Schema for file details."""

    file_id = fields.String(load_default=lambda: uuid.uuid4().hex)
    content_type = fields.String(load_default="unknown")
    file_name = fields.String(required=True)

    # measured in bytes (comes from stat() - st_size)
    file_size = fields.Integer(required=True)

    relative_path = fields.String(required=True)
    is_archive = fields.Boolean(load_default=False)
    is_dir = fields.Boolean(required=True)


class RenkuSyncSchema(Schema):
    """Parent schema for all Renku write operations."""

    remote_branch = fields.String()


class JobDetailsResponse(Schema):
    """Response schema for enqueued job."""

    job_id = fields.String()
    created_at = fields.DateTime()


class DelayedResponseRPC(JsonRPCResponse):
    """RPC response schema for project migrate."""

    result = fields.Nested(JobDetailsResponse)


class ErrorResponse(Schema):
    """Renku Service Error Response."""

    code = fields.Integer(required=True)
    userMessage = fields.String(required=True)
    devMessage = fields.String(required=True)
    userReference = fields.String()
    devReference = fields.String()
    sentry = fields.String()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""OLOS API integration."""

import datetime
import urllib
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional
from urllib import parse as urlparse
from uuid import UUID, uuid4

from renku.core import errors
from renku.core.config import get_value, set_value
from renku.core.dataset.providers.api import ExporterApi, ExportProviderInterface, ProviderApi, ProviderPriority
from renku.core.util import communication
from renku.domain_model.project_context import project_context

if TYPE_CHECKING:
    from renku.core.dataset.providers.models import ProviderParameter
    from renku.domain_model.dataset import Dataset, DatasetTag


class OLOSProvider(ProviderApi, ExportProviderInterface):
    """Provider for OLOS integration."""

    priority = ProviderPriority.HIGH
    name = "OLOS"
    is_remote = True

    def __init__(self, uri: str, is_doi: bool = False):
        super().__init__(uri=uri)

        self.is_doi = is_doi
        self._server_url = None

    @staticmethod
    def supports(uri):
        """Check if provider supports a given URI for importing."""
        return False

    @staticmethod
    def get_export_parameters() -> List["ProviderParameter"]:
        """Returns parameters that can be set for export."""
        from renku.core.dataset.providers.models import ProviderParameter

        return [ProviderParameter("dlcm-server", help="DLCM server base url.", type=str)]

    def get_exporter(
        self, dataset: "Dataset", *, tag: Optional["DatasetTag"], dlcm_server: Optional[str] = None, **kwargs
    ) -> "OLOSExporter":
        """Create export manager for given dataset."""

        def set_export_parameters():
            """Set and validate required parameters for exporting for a provider."""
            server = dlcm_server
            config_base_url = "server_url"
            if not server:
                server = get_value("olos", config_base_url)
            else:
                set_value("olos", config_base_url, server, global_only=True)

            if not server:
                raise errors.ParameterError("OLOS server URL is required.")

            self._server_url = server  # type: ignore

        set_export_parameters()
        return OLOSExporter(dataset=dataset, server_url=self._server_url)


class OLOSExporter(ExporterApi):
    """OLOS export manager."""

    def __init__(self, *, dataset, server_url=None):
        super().__init__(dataset)
        self._access_token = None
        self._server_url = server_url

    def set_access_token(self, access_token):
        """Set access token."""
        self._access_token = access_token

    def get_access_token_url(self):
        """Endpoint for creation of access token."""
        return urllib.parse.urljoin(self._server_url, "portal")

    def export(self, **kwargs):
        """Execute export process."""
        from renku.domain_model.dataset import get_file_path_in_dataset

        deposition = _OLOSDeposition(server_url=self._server_url, access_token=self._access_token)
        repository = project_context.repository

        metadata = self._get_dataset_metadata()
        metadata["organizationalUnitId"] = deposition.get_org_unit()
        deposition.create_dataset(metadata=metadata)

        with communication.progress("Uploading files ...", total=len(self.dataset.files)) as progressbar:
            for file in self.dataset.files:
                filepath = repository.copy_content_to_file(path=file.entity.path, checksum=file.entity.checksum)
                path_in_dataset = get_file_path_in_dataset(dataset=self.dataset, dataset_file=file)
                deposition.upload_file(full_path=filepath, path_in_dataset=path_in_dataset)
                progressbar.update()

        return deposition.deposited_at

    def _get_dataset_metadata(self):
        try:
            identifier = str(UUID(self.dataset.identifier, version=4))
        except ValueError:
            identifier = uuid4().hex

        metadata = {
            "publicationDate": datetime.date.today().isoformat(),
            "description": self.dataset.description,
            "identifier": identifier,
            "keywords": self.dataset.keywords,
            "title": self.dataset.title,
            "access": "CLOSED",
            "dataSensitivity": "CRIMSON",
            "year": datetime.datetime.today().year,
        }
        return metadata


class _OLOSDeposition:
    """OLOS record for deposit."""

    def __init__(
        self,
        *,
        access_token,
        server_url,
        dataset_pid=None,
        deposited_at=None,
        deposition_base_url=None,
        admin_base_url=None,
    ):
        self.access_token = access_token
        self.server_url = server_url
        self.dataset_pid = dataset_pid
        self.deposited_at = deposited_at
        self.deposition_base_url = deposition_base_url
        self.admin_base_url = admin_base_url

        self._get_base_urls()

    ORGANIZATIONAL_UNIT_PATH = "/authorized-organizational-units"
    DATASET_CREATE_PATH = "/deposits"
    FILE_UPLOAD_PATH = "/deposits/{deposit_id}/upload"
    MODULES_PATH = "administration/preservation-planning/modules"

    def _get_base_urls(self):
        """Get base urls for different endpoints."""
        url = self._make_url(self.server_url, api_path=self.MODULES_PATH)
        response = self._get(url=url)
        self._check_response(response)

        response_data = response.json()
        self.deposition_base_url = response_data["preingest"]
        self.admin_base_url = response_data["admin"]

    def get_org_unit(self):
        """Get the org units of the user."""
        url = self.admin_base_url + self.ORGANIZATIONAL_UNIT_PATH

        response = self._get(url=url)
        self._check_response(response)

        response_data = response.json()
        options = [(o["resId"], o["name"]) for o in response_data["_data"]]

        if not options:
            raise errors.ExportError("Couldn't export dataset, you aren't in any Organizational Unit.")

        if len(options) == 1:
            return options[0][0]
        ids, names = zip(*options)
        msg = "Organizational Unit to export to:\n"
        msg += "\n".join(f"{n} [{i}]" for i, n in enumerate(names, start=1))
        msg += "\n\n Organizational Unit"
        selection = communication.prompt(msg, type=int, default=1)

        return ids[selection - 1]

    def create_dataset(self, metadata):
        """Create a dataset in OLOS."""
        url = self.deposition_base_url + self.DATASET_CREATE_PATH

        response = self._post(url=url, json=metadata)
        self._check_response(response)

        response_data = response.json()

        self.dataset_pid = response_data["resId"]
        self.deposited_at = response_data["_links"]["self"]["href"]

        return response

    def upload_file(self, full_path, path_in_dataset: Path):
        """Upload a file to a previously-created dataset."""
        if self.dataset_pid is None:
            raise errors.ExportError("Dataset not created.")

        url = self.deposition_base_url + self.FILE_UPLOAD_PATH.format(deposit_id=self.dataset_pid)

        location = str(path_in_dataset.parent)

        if location == ".":
            location = ""

        params = {"folder": location, "category": "Primary", "type": "Derived"}

        files = {"file": (path_in_dataset.name, open(full_path, "rb"))}

        response = self._post(url=url, data=params, files=files)
        self._check_response(response)

        return response

    @staticmethod
    def _make_url(server_url, api_path, **query_params):
        """Create URL for creating a dataset."""
        url_parts = urlparse.urlparse(server_url)

        query_params_encoded = urllib.parse.urlencode(query_params)
        url_parts = url_parts._replace(path=api_path, query=query_params_encoded)
        return urllib.parse.urlunparse(url_parts)

    def _get(self, url):
        from renku.core.util import requests

        headers = {"Authorization": f"Bearer {self.access_token}"}
        try:
            return requests.get(url=url, headers=headers)
        except errors.RequestError as e:
            raise errors.ExportError("Cannot GET from remote server.") from e

    def _post(self, url, json=None, data=None, files=None):
        from renku.core.util import requests

        headers = {"Authorization": f"Bearer {self.access_token}"}
        try:
            return requests.post(url=url, json=json, data=data, files=files, headers=headers)
        except errors.RequestError as e:
            raise errors.ExportError("Cannot POST to remote server.") from e

    @staticmethod
    def _check_response(response):
        from renku.core.util import requests

        if len(response.history) > 0:
            raise errors.ExportError(
                f"Couldn't execute request to {response.request.url}, got redirected to {response.url}."
                "Maybe you mixed up http and https in the server url?"
            )

        try:
            requests.check_response(response=response)
        except errors.RequestError:
            json_res = response.json()
            raise errors.ExportError(
                "HTTP {} - Cannot export dataset: {}".format(
                    response.status_code, json_res["message"] if "message" in json_res else json_res["status"]
                )
            )



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""OS utility functions."""

import fnmatch
import glob
import hashlib
import io
import os
import re
import shutil
import subprocess
from pathlib import Path
from typing import Any, BinaryIO, Dict, Generator, List, Optional, Sequence, Union

from renku.core import errors

BLOCK_SIZE = 4096


def get_relative_path_to_cwd(path: Union[Path, str]) -> str:
    """Get a relative path to current working directory."""
    absolute_path = os.path.abspath(path)
    return os.path.relpath(absolute_path, os.getcwd())


def get_expanded_user_path(path: Union[Path, str]) -> str:
    """Expand the path if it starts with ``~``."""
    return "" if not path else os.path.expanduser(path)


def get_absolute_path(
    path: Union[Path, str], base: Optional[Union[Path, str]] = None, resolve_symlinks: bool = False, expand: bool = True
) -> str:
    """Return absolute normalized path.

    Args:
        path(Union[Path, str]): Path to get its absolute.
        base(Union[Path, str]): Base path to get absolute path from it.
        resolve_symlinks(bool): Whether to keep or resolve symlinks.
        expand(bool): Whether to expand ``~`` or not (Default value = True)

    Returns:
        str: Absolute path.
    """
    if expand:
        path = get_expanded_user_path(path)
    if base is not None:
        if expand:
            base = get_expanded_user_path(base)
        base = Path(base).resolve() if resolve_symlinks else os.path.abspath(base)
        path = os.path.join(base, path)

    if resolve_symlinks:
        return os.path.realpath(path)
    else:
        # NOTE: Do not use os.path.realpath or Path.resolve() because they resolve symlinks
        return os.path.abspath(path)


def get_safe_relative_path(path: Union[Path, str], base: Union[Path, str]) -> Path:
    """Return a relative path to the base and check path is within base with all symlinks resolved.

    NOTE: This is used to prevent path traversal attack.
    """
    try:
        base = Path(base).resolve()
        absolute_path = base / path
        return absolute_path.resolve().relative_to(base)
    except ValueError:
        raise ValueError(f"Path '{path}' is not with base directory '{base}'")


def get_relative_path(path: Union[Path, str], base: Union[Path, str], strict: bool = False) -> Optional[str]:
    """Return a relative path to the base if path is within base without resolving symlinks."""
    try:
        absolute_path = get_absolute_path(path=path, base=base)
        return str(Path(absolute_path).relative_to(base))
    except ValueError:
        if strict:
            raise errors.ParameterError(f"File {path} is not within path {base}")

    return None


def is_subpath(path: Union[Path, str], base: Union[Path, str]) -> bool:
    """Return True if path is within or same as base."""
    absolute_path = get_absolute_path(path=path)
    absolute_base = get_absolute_path(path=base)
    try:
        Path(absolute_path).relative_to(absolute_base)
    except ValueError:
        return False
    else:
        return True


def get_relative_paths(paths: Sequence[Union[Path, str]], base: Union[Path, str]) -> List[str]:
    """Return a list of paths relative to a base path."""
    relative_paths = []

    for path in paths:
        relative_path = get_relative_path(path=path, base=base)
        if relative_path is None:
            raise errors.ParameterError(f"Path '{path}' is not within base path '{base}'")

        relative_paths.append(relative_path)

    return relative_paths


def get_files(path: Path) -> Generator[Path, None, None]:
    """Return all files from a starting file/directory."""
    if not path.is_dir():
        yield path
    else:
        for subpath in path.rglob("*"):
            if not subpath.is_dir():
                yield subpath


def are_paths_related(a, b) -> bool:
    """Return True if paths are equal or one is the parent of the other."""
    common_path = os.path.commonpath((a, b))
    absolute_common_path = os.path.abspath(common_path)
    return absolute_common_path == os.path.abspath(a) or absolute_common_path == os.path.abspath(b)


def are_paths_equal(a: Union[Path, str], b: Union[Path, str]) -> bool:
    """Returns if two paths are the same."""
    # NOTE: The two paths should be identical; we don't consider the case where one is a sub-path of another
    return get_absolute_path(a) == get_absolute_path(b)


def is_path_empty(path: Union[Path, str]) -> bool:
    """Check if path contains files.

    :ref path: target path
    """
    subpaths = Path(path).glob("*")
    return not any(subpaths)


def create_symlink(target: Union[Path, str], symlink_path: Union[Path, str], overwrite: bool = True) -> None:
    """Create a symlink that points from symlink_path to target."""
    # NOTE: Don't resolve symlink path
    absolute_symlink_path = get_absolute_path(symlink_path)
    absolute_path = get_absolute_path(target, resolve_symlinks=True)

    Path(absolute_symlink_path).parent.mkdir(parents=True, exist_ok=True)

    try:
        if overwrite:
            delete_path(absolute_symlink_path)
        os.symlink(absolute_path, absolute_symlink_path)
    except OSError:
        raise errors.InvalidFileOperation(f"Cannot create symlink from '{symlink_path}' to '{target}'")


def delete_path(path: Union[Path, str]) -> None:
    """Delete a file/directory/symlink."""
    try:
        os.remove(path)
    except FileNotFoundError:
        pass
    except (PermissionError, IsADirectoryError, OSError):
        shutil.rmtree(path, ignore_errors=True)


def unmount_path(path: Union[Path, str]) -> None:
    """Unmount the given path and ignore all errors."""

    def execute_command(*command: str) -> bool:
        try:
            subprocess.run(command, check=True, text=True, capture_output=True)
        except (subprocess.CalledProcessError, FileNotFoundError):
            return False
        else:
            return True

    path = str(path)

    # NOTE: A symlink means that the path is not mounted itself but it's a link to a mount-point; just delete the link.
    if os.path.islink(path):
        os.remove(path)
        return

    # NOTE: ``fusermount`` is available on linux and ``umount`` is for macOS
    result = False
    if shutil.which("fusermount"):
        result = execute_command("fusermount", "-u", "-z", path)

    if not result:
        execute_command("umount", path)


def is_ascii(data):
    """Check if provided string contains only ascii characters."""
    return len(data) == len(data.encode())


def get_file_size(path: Union[Path, str], follow_symlinks: bool = True) -> Optional[int]:
    """Return size of a file in bytes."""
    path = Path(path).resolve() if follow_symlinks else Path(path)
    try:
        return path.stat().st_size
    except OSError:
        return None


def normalize_to_ascii(input_string, sep="-"):
    """Convert a string to only contain ASCII characters, with non-ASCII substring replaced with ``sep``."""
    replace_all = [sep, "_", "."]
    for replacement in replace_all:
        input_string = input_string.replace(replacement, " ")

    return (
        sep.join(
            [
                component
                for component in re.sub(r"[^a-zA-Z0-9_.-]+", " ", input_string).split(" ")
                if component and is_ascii(component)
            ]
        )
        .lower()
        .strip(sep)
    )


def delete_dataset_file(filepath: Union[Path, str], ignore_errors: bool = True, follow_symlinks: bool = False):
    """Remove a file/symlink and its pointer file (for external files)."""
    path = Path(filepath)
    link = None
    try:
        link = path.parent / os.readlink(path)
    except FileNotFoundError:
        if not ignore_errors:
            raise
        return
    except OSError:  # not a symlink but a normal file
        pass

    try:
        os.remove(path)
    except OSError:
        if not ignore_errors:
            raise

    if follow_symlinks and link:
        try:
            os.remove(link)
        except FileNotFoundError:
            pass


def hash_file(path: Union[Path, str], hash_type: str = "sha256") -> Optional[str]:
    """Calculate the sha256 hash of a file."""
    if not os.path.exists(path):
        return None

    with open(path, "rb") as f:
        return hash_file_descriptor(f, hash_type)


def hash_string(content: str, hash_type: str = "sha256") -> str:
    """Hash a string."""
    content_bytes = content.encode("utf-8")
    file = io.BytesIO(content_bytes)

    return hash_file_descriptor(file, hash_type)


def hash_file_descriptor(file: BinaryIO, hash_type: str = "sha256") -> str:
    """Hash content of a file descriptor."""
    hash_type = hash_type.lower()
    assert hash_type in ("sha256", "md5")

    hash_value = hashlib.sha256() if hash_type == "sha256" else hashlib.md5()  # nosec

    for byte_block in iter(lambda: file.read(BLOCK_SIZE), b""):
        hash_value.update(byte_block)

    return hash_value.hexdigest()


def safe_read_yaml(path: Union[Path, str]) -> Dict[str, Any]:
    """Parse a YAML file.

    Returns:
        In case of success a dictionary of the YAML's content, otherwise raises a ParameterError exception.
    """
    try:
        from renku.core.util import yaml as yaml

        return yaml.read_yaml(path)
    except Exception as e:
        raise errors.ParameterError(e)


def matches(path: Union[Path, str], pattern: str) -> bool:
    """Check if a path matched a given pattern."""
    pattern = pattern.rstrip(os.sep)

    path = Path(path)
    paths = [path] + list(path.parents)[:-1]

    for parent in paths:
        if fnmatch.fnmatch(str(parent), pattern):
            return True

    return False


def expand_directories(paths):
    """Expand directory with all files it contains."""
    processed_paths = set()
    for path in paths:
        for matched_path in glob.iglob(str(path), recursive=True):
            if matched_path in processed_paths:
                continue
            path_ = Path(matched_path)
            if path_.is_dir():
                for expanded in path_.rglob("*"):
                    processed_paths.add(str(expanded))
                    yield str(expanded)
            else:
                processed_paths.add(matched_path)
                yield matched_path


UNITS = {
    "b": 1,
    "kb": 1000,
    "mb": 1000**2,
    "gb": 1000**3,
    "tb": 1000**4,
    "m": 1000**2,
    "g": 1000**3,
    "t": 1000**4,
    "p": 1000**5,
    "e": 1000**6,
    "z": 1000**7,
    "y": 1000**8,
    "ki": 1024,
    "mi": 1024**2,
    "gi": 1024**3,
    "ti": 1024**4,
    "pi": 1024**5,
    "ei": 1024**6,
    "zi": 1024**7,
    "yi": 1024**8,
}


def parse_file_size(size_str):
    """Parse a human readable file size to bytes."""
    res = re.search(r"([0-9.]+)([a-zA-Z]{1,2})", size_str)
    if not res or res.group(2).lower() not in UNITS:
        raise ValueError(
            "Supplied file size does not contain a unit. " "Valid units are: {}".format(", ".join(UNITS.keys()))
        )

    value = float(res.group(1))
    unit = UNITS[res.group(2).lower()]

    return int(value * unit)


def bytes_to_unit(size_in_bytes, unit: str) -> Optional[float]:
    """Return size in the provided unit."""
    unit = unit.lower()
    if unit not in UNITS:
        raise ValueError(f"Invalid unit '{unit}'. Valid units are: [{', '.join(UNITS)}]")
    return None if size_in_bytes is None else size_in_bytes / UNITS[unit]



#
# Copyright 2018-2023 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku CLI commands for handling of Knowledge Graph data.

Description
~~~~~~~~~~~

Export part or all of the Renku Knowledge Graph metadata for the
current project.

By default, this will export the metadata created in the last commit in the
project.
If that commit was not a ``renku`` command that creates metadata, it will
produce no output.

Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.graph:graph
   :prog: renku graph
   :nested: full

Examples
~~~~~~~~

Here we create a new dataset and then use ``renku graph export`` to export the
created metadata as JSON-LD, the default format.

.. code-block:: console

   $ renku dataset create my-dataset
   OK
   $ renku graph export
    [
        {
            "@id": "https://localhost/datasets/850e74d6c0204e8c923457a1b9ce52d8",
            "@type": [
            "http://schema.org/Dataset",
            "http://www.w3.org/ns/prov#Entity"
            ],

            [... many more lines ...]

        }
    ]

If you want the Knowledge Graph data for the whole project, you can use
``renku graph export --full`` (``--full`` is the default). Alternatively,
you can get data for a single commit by using ``renku graph export
--revision <git commit sha>`` or by specifying a range of commits like
``renku graph export --revision sha1..sha2``.

``renku graph export`` currently supports various formats for export, such as
``json-ld``, ``rdf``, ``nt`` (for triples) and ``dot`` (for GraphViz graphs),
which can be specified using the ``--format`` option. For instance,

.. code-block:: console

   $ renku graph export --full --format dot | dot -Tpng -o my_graph.png

would produce a PNG image of the whole Knowledge Graph for the project.

To run validation on the generated output, you can pass the ``--strict``
option, which will check that all the nodes and properties in the graph are
correct and that there isn't anything missing.

"""

import click

from renku.ui.cli.utils.click import CaseInsensitiveChoice

GRAPH_FORMATS = {
    "jsonld": "jsonld",
    "json-ld": "jsonld",
    "nt": "nt",
    "rdf": "rdf",
    "dot": "dot",
    "dot-landscape": "dot-landscape",
    "dot-debug": "dot-debug",
}


@click.group()
def graph():
    """Graph commands."""


@graph.command()
@click.option(
    "--format",
    type=CaseInsensitiveChoice(list(GRAPH_FORMATS.keys())),
    default="json-ld",
    help="Choose an output format.",
)
@click.option(
    "--revision",
    type=str,
    default=None,
    help="Limit graph to changes done in revision (or range of revisions like 'A..B').",
)
@click.option("-f", "--full", is_flag=True, help="Generate full graph for project (default). Overrides --revision.")
@click.option("--strict", is_flag=True, default=False, help="Validate triples before output.")
@click.option(
    "--no-indent", is_flag=True, default=False, help="Format without indentation/pretty-printing (only for JSON-LD)."
)
def export(format, revision, full, strict, no_indent):
    r"""Export Renku graph metadata for project."""
    from renku.command.graph import export_graph_command
    from renku.command.view_model.graph import DotFormat
    from renku.ui.cli.utils.callback import ClickCallback

    if full:
        revision = None

    communicator = ClickCallback()
    result = (
        export_graph_command()
        .with_communicator(communicator)
        .build()
        .execute(format=format, strict=strict, revision_or_range=revision)
    )
    format = GRAPH_FORMATS[format]

    if format == "jsonld":
        result = result.output.as_jsonld_string(indentation=None if no_indent else 2)
    elif format == "rdf":
        result = result.output.as_rdf_string()
    elif format == "nt":
        result = result.output.as_nt_string()
    elif format == "dot":
        result = result.output.as_dot_string(format=DotFormat.FULL)
    elif format == "dot-landscape":
        result = result.output.as_dot_string(format=DotFormat.FULL_LANDSCAPE)
    elif format == "dot-debug":
        result = result.output.as_dot_string(format=DotFormat.DEBUG)
    else:
        raise NotImplementedError(f"Format {format} not supported for graph export.")

    click.echo(result)



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service template serializers."""

from marshmallow import Schema, fields

from renku.ui.service.serializers.rpc import JsonRPCResponse


class ManifestTemplateSchema_1_5(Schema):
    """Manifest template schema."""

    description = fields.String(required=True, metadata={"description": "Description of the template"})
    folder = fields.String(required=True, metadata={"description": "Folder the template resides in"})
    name = fields.String(required=True, metadata={"description": "Name of the template"})
    variables = fields.Dict(
        load_default={}, metadata={"description": "Dictionary of values that can be set on this template"}
    )
    icon = fields.String(
        load_default=None, metadata={"description": "base64 encoded icon for the template in PNG format"}
    )


class ManifestTemplatesResponse_1_5(Schema):
    """Manifest templates response."""

    templates = fields.List(fields.Nested(ManifestTemplateSchema_1_5), required=True)


class ManifestTemplatesResponseRPC_1_5(JsonRPCResponse):
    """RPC schema for listing manifest templates."""

    result = fields.Nested(ManifestTemplatesResponse_1_5)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent a group of run templates."""

import copy
from collections import defaultdict
from datetime import datetime
from typing import Callable, Dict, List, Optional, Set, Tuple, Union, cast
from uuid import uuid4

from renku.core import errors
from renku.core.util.datetime8601 import local_now
from renku.domain_model.provenance.agent import Person
from renku.domain_model.provenance.annotation import Annotation
from renku.domain_model.workflow.parameter import (
    CommandInput,
    CommandOutput,
    CommandParameter,
    CommandParameterBase,
    ParameterLink,
    ParameterMapping,
)
from renku.domain_model.workflow.plan import MAX_GENERATED_NAME_LENGTH, AbstractPlan, Plan


class CompositePlan(AbstractPlan):
    """A plan containing child plans."""

    annotations: List[Annotation] = list()

    def __init__(
        self,
        *,
        derived_from: Optional[str] = None,
        description: Optional[str] = None,
        id: str,
        date_created: Optional[datetime] = None,
        date_modified: Optional[datetime] = None,
        date_removed: Optional[datetime] = None,
        keywords: Optional[List[str]] = None,
        links: Optional[List[ParameterLink]] = None,
        mappings: Optional[List[ParameterMapping]] = None,
        name: str,
        plans: List[Union["CompositePlan", Plan]],
        project_id: Optional[str] = None,
        annotations: Optional[List[Annotation]] = None,
        creators: Optional[List[Person]] = None,
    ):
        super().__init__(
            derived_from=derived_from,
            description=description,
            id=id,
            date_created=date_created,
            date_modified=date_modified,
            date_removed=date_removed,
            keywords=keywords,
            name=name,
            project_id=project_id,
            creators=creators,
        )
        self.annotations: List[Annotation] = annotations or []

        self.plans: List[Union["CompositePlan", Plan]] = plans
        self.mappings: List[ParameterMapping] = mappings or []
        self.links: List[ParameterLink] = links or []

    def _find_existing_mapping(
        self, targets: List[CommandParameterBase]
    ) -> Dict[CommandParameterBase, List[ParameterMapping]]:
        """Find if any mapping already exists for any of the targets."""
        existing = defaultdict(list)
        for target in targets:
            found = [m for m in self.mappings if target in m.mapped_parameters]

            if found:
                existing[target].extend(found)

        return dict(existing)

    def is_equal_to(self, other: "CompositePlan") -> bool:
        """Return true if plan hasn't changed from the other plan."""

        def are_equal_with_order(values, other_values):
            return len(values) == len(other_values) and all(s.is_equal_to(o) for s, o in zip(values, other_values))

        def are_equal(values, other_values):
            return len(values) == len(other_values) and set(values) == set(other_values)

        # TODO: Include ``annotations``, ``mappings``, and ``links`` if they are added to the workflow definition file
        return (
            self.name == other.name
            and self.description == other.description
            and self.project_id == other.project_id
            and are_equal(self.keywords, other.keywords)
            and are_equal(self.creators, other.creators)
            and are_equal_with_order(self.plans, other.plans)
        )

    def set_mappings_from_strings(self, mapping_strings: List[str]) -> None:
        """Set mappings by parsing mapping strings."""
        for mapping_string in mapping_strings:
            name, targets = mapping_string.split("=", maxsplit=1)

            if not targets:
                # remove mapping that has no target
                for mapping in self.mappings:
                    if mapping.name == name:
                        self.mappings.remove(mapping)
                        break
                else:
                    raise errors.MappingNotFoundError(mapping=name, workflow=self.name)
                continue

            target_names = targets.split(",")

            target_params = []

            for target in target_names:
                target_params.append(self.resolve_mapping_path(target)[0])

            existing = self._find_existing_mapping(target_params)

            if existing:
                raise errors.MappingExistsError(
                    [k.name + ": " + ", ".join(m.name for m in v) for k, v in existing.items()]
                )

            self.add_mapping(
                ParameterMapping(
                    name=name,
                    mapped_parameters=target_params,
                    id=ParameterMapping.generate_id(plan_id=self.id, postfix=name),
                    default_value=None,
                )
            )

    def add_mapping(self, mapping: ParameterMapping) -> None:
        """Add a mapping to this run."""
        existing: Optional[ParameterMapping] = next((m for m in self.mappings if mapping.name == m.name), None)

        if not existing:
            self.mappings.append(mapping)
            return

        new = set(mapping.mapped_parameters) - set(existing.mapped_parameters)

        if new:
            existing.mapped_parameters.extend(new)

    def set_links_from_strings(self, link_strings: List[str]) -> None:
        """Set links between parameters of child steps."""
        for link_string in link_strings:
            source, sinks = link_string.split("=", maxsplit=1)

            source_param, _ = self.resolve_mapping_path(source)

            if isinstance(source_param, ParameterMapping):
                sources = list(source_param.leaf_parameters)
            else:
                sources = [source_param]

            sinks_names = sinks.split(",")

            sink_params: List[CommandParameterBase] = []

            for sink in sinks_names:
                resolved_sink, _ = self.resolve_mapping_path(sink)

                if isinstance(resolved_sink, ParameterMapping):
                    # NOTE: can't link between mappings, only the parameters they map to
                    sink_params.extend(resolved_sink.leaf_parameters)
                else:
                    sink_params.append(resolved_sink)

            for link_source in sources:
                self.add_link(link_source, sink_params)

    def add_link(self, source: CommandParameterBase, sinks: List[CommandParameterBase]) -> None:
        """Validate and add a ParameterLink."""
        if not source:
            raise errors.ParameterLinkError("Parameter Link has no source.")

        if not sinks:
            raise errors.ParameterLinkError("Parameter Link has no sink(s).")

        if isinstance(source, (CommandInput, CommandParameter)):
            # TODO: Change this once parameters can be calculated and serve as output
            raise errors.ParameterLinkError(
                f"A parameter link can't originate in a command input or command parameter: {source.name}"
            )

        for sink in sinks:
            if isinstance(sink, CommandOutput):
                raise errors.ParameterLinkError(f"A parameter link can't end in a command output: '{sink.name}'.")

            source_wf = self.find_parameter_workflow(source)
            if source_wf is None:
                raise errors.ParameterLinkError(f"Couldn't find parameter {source}")

            sink_wf = self.find_parameter_workflow(sink)
            if sink_wf is None:
                raise errors.ParameterLinkError(f"Couldn't find parameter {sink}")

            if source_wf.find_parameter(sink) or sink_wf.find_parameter(source):
                raise errors.ParameterLinkError(
                    f"Parameter links have to link between different workflows, source '{source.name}' and "
                    f"sink '{sink.name}' are on the same workflow."
                )

            existing = list(self.find_link_by_target(cast(Union[CommandInput, CommandParameter], sink)))

            if existing:
                _, existing_wfs = zip(*existing)
                existing_wfs = ", ".join(w.name for w in existing_wfs)

                raise errors.ParameterLinkError(
                    f"Parameter link to sink '{sink.name}' already exists in workflows: {existing_wfs}"
                )

        self.links.append(ParameterLink(source=source, sinks=sinks, id=ParameterLink.generate_id(self.id)))

    def find_parameter(self, parameter: CommandParameterBase):
        """Check if a parameter exists on this run or one of its children."""
        if parameter in self.mappings:
            return True

        for plan in self.plans:
            if plan.find_parameter(parameter):
                return True

        return False

    def get_parameter_path(self, parameter: CommandParameterBase):
        """Get the path to a parameter inside this plan."""
        if parameter in self.mappings:
            return [self]

        for plan in self.plans:
            path = plan.get_parameter_path(parameter)
            if path is not None:
                return [self] + path

        return None

    def get_parameter_by_id(self, parameter_id: str) -> Optional[CommandParameterBase]:
        """Get a parameter on this plan by id."""
        mapping = next((p for p in self.mappings if parameter_id == p.id), None)

        if mapping:
            return mapping

        for plan in self.plans:
            parameter = plan.get_parameter_by_id(parameter_id)

            if parameter:
                return parameter

        return None

    def find_parameter_workflow(self, parameter: CommandParameterBase) -> Optional[AbstractPlan]:
        """Return the workflow a parameter belongs to."""
        if parameter in self.mappings:
            return self

        for plan in self.plans:
            found = plan.find_parameter_workflow(parameter)
            if found:
                return found

        return None

    def find_link_by_target(self, target: Union[CommandInput, CommandParameter]):
        """Find a link on this or a child workflow that has target as a sink."""
        for link in self.links:
            if target in link.sinks:
                yield link, self

        for plan in self.plans:
            if isinstance(plan, CompositePlan):
                yield plan.find_link_by_target(target)

    def _map_all(self, selector: Callable[[Plan], List[CommandParameterBase]]) -> None:
        """Automatically map all base parameters matched by selection_lambda on this run."""

        for step in self.plans:
            if not isinstance(step, Plan):
                continue

            params = selector(step)

            existing = self._find_existing_mapping(params)
            non_mapped: Set[CommandParameterBase] = set(params) - existing.keys()

            for param in non_mapped:
                name = f"{step.name}_{param.name}"
                self.add_mapping(
                    ParameterMapping(
                        name=name,
                        mapped_parameters=[param],
                        id=ParameterMapping.generate_id(
                            plan_id=self.id,
                            postfix=name,
                        ),
                        default_value=param.default_value,
                    )
                )

    def map_all_inputs(self) -> None:
        """Map all unmapped inputs from child steps to the parent."""
        self._map_all(lambda x: cast(List[CommandParameterBase], x.inputs))

    def map_all_outputs(self) -> None:
        """Map all unmapped outputs from child steps to the parent."""
        self._map_all(lambda x: cast(List[CommandParameterBase], x.outputs))

    def map_all_parameters(self) -> None:
        """Map all unmapped parameters from child steps to the parent."""
        self._map_all(lambda x: cast(List[CommandParameterBase], x.parameters))

    def set_mapping_defaults(self, default_strings: List[str]) -> None:
        """Set default value based on a default specification string."""
        for default_string in default_strings:
            target, value = default_string.split("=", maxsplit=1)

            mapping, _ = self.resolve_mapping_path(target)

            mapping.default_value = value

    def set_mapping_descriptions(self, mapping_descriptions: List[str]) -> None:
        """Set descriptions for mappings."""

        for mapping_description in mapping_descriptions:
            target, value = mapping_description.split("=", maxsplit=1)

            mapping = self.resolve_direct_reference(target)
            mapping.description = value.strip(' "')

    def resolve_mapping_path(self, mapping_path: str) -> Tuple[CommandParameterBase, Union["CompositePlan", Plan]]:
        """Resolve a mapping path to its reference parameter."""

        parts = mapping_path.split(".", maxsplit=1)

        if len(parts) == 1:
            return self.resolve_direct_reference(parts[0]), self

        prefix, suffix = parts

        if prefix.startswith("@step"):
            # NOTE: relative reference
            try:
                workflow = self.plans[int(prefix[5:]) - 1]
            except (ValueError, IndexError):
                raise errors.ParameterNotFoundError(mapping_path, self.name)
            if isinstance(workflow, CompositePlan):
                return workflow.resolve_mapping_path(suffix)
            else:
                return workflow.resolve_direct_reference(suffix), workflow

        for workflow in self.plans:
            if workflow.name == prefix:
                return workflow.resolve_mapping_path(suffix)

        raise errors.ParameterNotFoundError(mapping_path, self.name)

    def resolve_direct_reference(self, reference: str) -> CommandParameterBase:
        """Resolve a direct parameter reference."""
        if reference.startswith("@mapping"):
            try:
                return self.mappings[int(reference[8:]) - 1]
            except (ValueError, IndexError):
                raise errors.ParameterNotFoundError(reference, self.name)

        for mapping in self.mappings:
            if mapping.name == reference:
                return mapping

        raise errors.ParameterNotFoundError(reference, self.name)

    def _get_default_name(self) -> str:
        return uuid4().hex[:MAX_GENERATED_NAME_LENGTH]

    def derive(self, creator: Optional[Person] = None) -> "CompositePlan":
        """Create a new ``CompositePlan`` that is derived from self."""
        derived = copy.copy(self)
        derived.derived_from = self.id
        derived.date_modified = local_now()
        derived.plans = self.plans.copy()
        derived.mappings = self.mappings.copy()
        derived.links = self.links.copy()
        derived.assign_new_id()

        if creator and hasattr(creator, "email") and not any(c for c in self.creators if c.email == creator.email):
            self.creators.append(creator)

        return derived

    def is_derivation(self) -> bool:
        """Return if an ``CompositePlan`` has correct derived_from."""
        return self.derived_from is not None and self.id != self.derived_from



#
# Copyright 2017-2023 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Renku API Dataset.

Dataset class allows listing datasets and files inside a Renku project and
accessing their metadata.

To get a list of available datasets in a Renku project use ``list`` method:

.. code-block:: python

    from renku.api import Dataset

    datasets = Dataset.list()

You can then access metadata of a dataset like ``name``, ``title``,
``keywords``, etc. To get the list of files inside a dataset use ``files``
property:

.. code-block:: python

    for dataset_file in dataset.files:
        print(dataset_file.path)

"""

from operator import attrgetter
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional

from renku.domain_model import dataset as core_dataset
from renku.domain_model.project_context import has_graph_files
from renku.infrastructure.gateway.dataset_gateway import DatasetGateway
from renku.ui.api.util import ensure_project_context

if TYPE_CHECKING:
    from renku.ui.api import Project


class Dataset:
    """API Dataset model."""

    _ATTRIBUTES = [
        "creators",
        "date_created",
        "date_published",
        "description",
        "keywords",
        "license",
        "name",
        "title",
        "url",
        "version",
    ]

    def __init__(self):
        self._dataset = None

        for name in self._ATTRIBUTES:
            setattr(self, name, None)
        self._files = []

    @classmethod
    def _from_dataset(cls, dataset: core_dataset.Dataset) -> "Dataset":
        """Create an instance from Dataset metadata.

        Args:
            dataset(core_dataset.Dataset): The core dataset to wrap.

        Returns:
            Dataset: An API ``Dataset`` wrapping a core dataset.
        """
        self = cls()
        self._dataset = dataset
        self._files = [DatasetFile._from_dataset_file(f) for f in self._dataset.files]

        return self

    @staticmethod
    @ensure_project_context
    def list(project: "Project") -> List["Dataset"]:
        """List all datasets in a project.

        Args:
            project: The current project

        Returns:
            List["Dataset"]: A list of all datasets in the supplied project.
        """
        if not project.repository or not has_graph_files():
            return []

        dataset_gateway = DatasetGateway()

        return [Dataset._from_dataset(d) for d in dataset_gateway.get_all_active_datasets()]

    def __getattribute__(self, name):
        dataset = object.__getattribute__(self, "_dataset")
        if dataset is not None and name in Dataset._ATTRIBUTES:
            return getattr(dataset, name)

        return object.__getattribute__(self, name)

    @property
    def files(self):
        """Return list of existing files."""
        return self._files


class DatasetFile:
    """API DatasetFile model."""

    _ATTRIBUTES = ["date_added", "name", "path", "entity"]

    _dataset_file: core_dataset.DatasetFile
    full_path: Optional[Path] = None

    @classmethod
    @ensure_project_context
    def _from_dataset_file(cls, dataset_file: core_dataset.DatasetFile, project: "Project"):
        """Create an instance from Dataset metadata.

        Args:
            dataset_file(core_dataset.DatasetFile): The ``DatasetFile`` to wrap.
            project: The current project.

        Returns:
            An API ``DatasetFile`` wrapping a core dataset file.
        """
        self = cls()
        self._dataset_file = dataset_file

        self.full_path = project.path / dataset_file.entity.path

        return self

    def __getattribute__(self, name):
        dataset_file = object.__getattribute__(self, "_dataset_file")
        if dataset_file is not None and name in DatasetFile._ATTRIBUTES:
            if name == "path":
                name = "entity.path"
            getter = attrgetter(name)
            return getter(dataset_file)

        return object.__getattribute__(self, name)



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service headers serializers."""
import base64
import binascii
import os

import jwt
from marshmallow import Schema, ValidationError, fields, post_load, pre_load
from werkzeug.utils import secure_filename

JWT_TOKEN_SECRET = os.getenv("RENKU_JWT_TOKEN_SECRET", "bW9menZ3cnh6cWpkcHVuZ3F5aWJycmJn")


def decode_b64(value):
    """Decode base64 values or return raw value."""
    try:
        decoded = base64.b64decode(value, validate=True)
        return decoded.decode("utf-8")
    except binascii.Error:
        return value


def encode_b64(value):
    """Encode value to base64."""
    if isinstance(value, str):
        value = bytes(value, "utf-8")

    return base64.b64encode(value).decode("utf-8")


class UserIdentityToken(Schema):
    """User identity token schema."""

    jti = fields.String()
    exp = fields.Integer()
    nbf = fields.Integer()
    iat = fields.Integer()
    iss = fields.String()
    aud = fields.List(fields.String())
    sub = fields.String()
    typ = fields.String()
    azp = fields.String()
    nonce = fields.String()
    auth_time = fields.Integer()
    session_state = fields.String()
    acr = fields.String()
    email_verified = fields.Boolean()
    preferred_username = fields.String()
    given_name = fields.String()
    family_name = fields.String()

    email = fields.String(required=True)
    name = fields.String(required=True)
    user_id = fields.String()  # INFO: Generated post load.

    @post_load
    def set_user_id(self, data, **kwargs):
        """Sets users id."""
        data["user_id"] = encode_b64(secure_filename(data["sub"]))
        return data


class RenkuHeaders:
    """Renku headers support."""

    @staticmethod
    def decode_token(token):
        """Extract authorization token."""
        components = token.split(" ")

        rfc_compliant = token.lower().startswith("bearer")
        rfc_compliant &= len(components) == 2

        if not rfc_compliant:
            raise ValidationError("authorization value contains invalid value")

        return components[-1]

    @staticmethod
    def decode_user(data):
        """Extract renku user from a JWT."""
        decoded = jwt.decode(data, JWT_TOKEN_SECRET, algorithms=["HS256"], audience="renku")
        return UserIdentityToken().load(decoded)

    @staticmethod
    def reset_old_headers(data):
        """Process old version of old headers."""
        # TODO: This should be removed once support for them is phased out.
        if "renku-user-id" in data:
            data.pop("renku-user-id")

        if "renku-user-fullname" in data and "renku-user-email" in data:
            renku_user = {
                "aud": ["renku"],
                "name": decode_b64(data.pop("renku-user-fullname")),
                "email": decode_b64(data.pop("renku-user-email")),
            }
            renku_user["sub"] = renku_user["email"]
            data["renku-user"] = jwt.encode(renku_user, JWT_TOKEN_SECRET, algorithm="HS256")

        return data


class IdentityHeaders(Schema):
    """User identity schema."""

    @pre_load
    def set_fields(self, data, **kwargs):
        """Set fields for serialization."""
        # NOTE: We don't process headers which are not meant for determining identity.
        # TODO: Remove old headers support once support for them is phased out.
        old_keys = ["renku-user-id", "renku-user-fullname", "renku-user-email"]
        expected_keys = old_keys + [field.data_key for field in self.fields.values()]

        data = {key.lower(): value for key, value in data.items() if key.lower() in expected_keys}
        data = RenkuHeaders.reset_old_headers(data)

        return data

    @post_load
    def set_user(self, data, **kwargs):
        """Extract user object from a JWT."""
        result = {}

        if "auth_token" in data:
            result["token"] = RenkuHeaders.decode_token(data["auth_token"])

        if data and "user_token" in data:
            user = RenkuHeaders.decode_user(data["user_token"])
            result["fullname"] = user.pop("name")
            result["email"] = user.pop("email")
            result["user_id"] = user.pop("user_id")

        return result


class RequiredIdentityHeaders(IdentityHeaders):
    """Identity schema for required headers."""

    user_token = fields.String(required=True, data_key="renku-user")
    auth_token = fields.String(required=True, data_key="authorization")


class OptionalIdentityHeaders(IdentityHeaders):
    """Identity schema for optional headers."""

    user_token = fields.String(data_key="renku-user")
    auth_token = fields.String(data_key="authorization")



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Manage the set of execution templates created by the ``renku run`` command.

Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.workflow:workflow
   :prog: renku workflow
   :nested: full

Description
~~~~~~~~~~~

Renku records two different kinds of metadata when a workflow is executed,
``Run`` and ``Plan``.
Plans describe a recipe for a command. They function as a template that
can be used directly or combined with other workflow templates to create more
complex recipes.
These Plans can be run in various ways, on creation with ``renku run``,
doing a ``renku rerun`` or ``renku update`` or manually using ``renku workflow
execute``.

Each time a ``Plan`` is run, we track that instance of it as a ``Run``.
Runs track workflow execution through time. They track which Plan was
run, at what time, with which specific values. This gives an insight into what
were the steps taken in a repository, how they were taken and what results they
produced.

The ``renku workflow`` group of commands contains most of the commands used
to interact with Plans and Runs

Working with Plans
~~~~~~~~~~~~~~~~~~

Listing Plans
*************

.. image:: ../../_static/asciicasts/list_plans.delay.gif
   :width: 850
   :alt: List Plans

.. code-block:: console

    $ renku workflow ls
    ID                                       NAME
    ---------------------------------------  ---------------
    /plans/11a3702184394b93ac422df760e40999  cp-B-C-ca4da
    /plans/96642cac86d9435e8abce2384f8618b9  cat-A-C-fa017
    /plans/96c70626575c41c5a13853b070eaaaf5  my-other-run
    /plans/9a0961844fcc46e1816fde00f57e24a8  my-run

Each entry corresponds to a recorded Plan/workflow template. You can also
show additional columns using the ``--columns`` parameter, which takes any
combination of values from ``id``, ``name``, ``keywords`` and ``description``.

.. cheatsheet::
   :group: Workflows
   :command: $ renku workflow ls
   :description: List Plans (workflow templates).
   :target: rp

Showing Plan Details
********************

.. image:: ../../_static/asciicasts/show_plan.delay.gif
   :width: 850
   :alt: Show Plan

You can see the details of a plan by using ``renku workflow show``:

.. code-block:: console

    $ renku workflow show my-run
    Id: /plans/9a0961844fcc46e1816fde00f57e24a8
    Name: run1
    Command: cp A B
    Success Codes:
    Inputs:
            - input-1:
                    Default Value: A
                    Position: 1
    Outputs:
            - output-2:
                    Default Value: B
                    Position: 2

This shows the unique Id of the Plan, its name, the full command of the Plan
if it was run without any modifications (more on that later), which exit codes
should be considered successful executions (defaults to ``0``) as well as its
inputs, outputs and parameters.

.. cheatsheet::
   :group: Workflows
   :command: $ renku workflow show <name>
   :description: Show details for Plan <name>.
   :target: rp

Executing Plans
***************

.. image:: ../../_static/asciicasts/execute_plan.delay.gif
   :width: 850
   :alt: Execute Plans

Plans can be executed using ``renku workflow execute``. They can be run as-is
or their parameters can be modified as needed. Renku has a plugin architecture
to allow execution using various execution backends.

.. code-block:: console

    $ renku workflow execute --provider cwltool --set input-1=file.txt my-run

Parameters can be set using the ``--set`` keyword or by specifying them in a
values YAML file and passing that using ``--values``. In case of passing a file,
for a composite workflow like:

.. code-block:: console

    $ renku run --name train -- python train.py --lr=0.1 --gamma=0.5 --output=result.csv
    $ renku run --name eval -- python eval.py --image=graph.png --data=result.csv
    $ renku workflow compose --map learning_rate=train.lr --map graph=eval.image

the YAML file could look like:

.. code-block:: yaml

    # composite (mapped) parameters
    learning_rate: 0.9
    graph: overview.png
    train: # child workflow name
        # child workflow parameters
        gamma: 1.0

Which would rerun the two steps but with ``lr`` set to ``0.9``, ``gamma`` set to ``1.0``
and the output saved under ``overview.png``.

Note that this would be the same as using:

.. code-block:: yaml

    train:
        lr: 0.9
        gamma: 1.0
    eval:
        image: overview.png

For a regular (non-composite) workflow it is enough to just specify key-value pairs like:

.. code-block:: yaml

    lr: 0.9
    gamma: 1.0

In addition to being passed on the command line and being available to
``renku.ui.api.*`` classes in Python scripts, parameters are also set as
environment variables when executing the command, in the form of
``RENKU_ENV_<parameter name>``.

Provider specific settings can be passed as file using the ``--config`` parameter.

.. cheatsheet::
   :group: Workflows
   :command: $ renku workflow execute --provider <provider> [--set
             <param-name>=<value>...] <name>
   :description: Execute a Plan using <provider> as a backend, overriding
                 parameter <param-name>'s value.
   :target: rp

In some cases it may be desirable to avoid updating the renku metadata
and to avoid committing this and any other change in the repository when a workflow
is executed. If this is the case then you can pass the ``--skip-metadata-update``
flag to ``renku workflow execute``.

Iterate Plans
*************

.. image:: ../../_static/asciicasts/iterate_plan.gif
   :width: 850
   :alt: Iterate Plans

For executing a Plan with different parametrization ``renku workflow iterate``
could be used. This sub-command is basically conducting a 'grid search'-like
execution of a Plan, with parameter-sets provided by the user.

.. code-block:: console

    $ renku workflow iterate --map parameter-1=[1,2,3] \
            --map parameter-2=[10,20] my-run

The set of possible values for a parameter can be given by ``--map`` command
line argument or by specifying them in a values YAML file and passing that
using ``--mapping``. Content of the mapping file for the above example
should be:

.. code-block:: yaml

    parameter-1: [1,2,3]
    parameter-2: [10,20]

.. cheatsheet::
   :group: Workflows
   :command: $ renku workflow iterate [--map <param-name>=[value,value,...]]
             <name>
   :description: Repeatedly execute a Plan, taking values from the list
                 specified with --map.
   :target: rp

By default ``renku workflow iterate`` will execute all the combination of the
given parameters' list of possible values. Sometimes it is desired that instead
of all the combination of possible values, a specific tuple of values are
executed. This could be done by marking the parameters that should be bound
together with the ``@tag`` suffix in their names.

.. code-block:: console

    $ renku workflow iterate --map parameter-1@tag1=[1,2,3] \
            --map parameter-2@tag1=[10,5,30] my-run

This will result in only three distinct execution of the ``my-run`` Plan,
with the following parameter combinations: ``[(1,10), (2,5), (3,30)]``. It is
important to note that parameters that have the same tag, should have the same
number of possible values, i.e. the values list should have the same length.

There's a special template variable for parameter values ``{iter_index}``, which
can be used to mark each iteration's index in a value of a parameter. The template
variable is going to be substituted with the iteration index (0, 1, 2, ...).

.. code-block:: console

    $ renku workflow iterate --map parameter-1=[10,20,30] \
            --map output=output_{iter_index}.txt my-run

This would execute ``my-run`` three times, where ``parameter-1`` values would be
``10``, ``20`` and ``30`` and the producing output files ``output_0.txt``,
``output_1.txt`` and ``output_2.txt`` files in this order.

In some cases it may be desirable to avoid updating the renku metadata
and to avoid committing this and any other change in the repository when a workflow
is iterated through. If this is the case then you can pass the ``--skip-metadata-update``
flag to ``renku workflow iterate``.

Exporting Plans
***************

You can export a Plan to a number of different workflow languages, such as CWL
(Common Workflow Language) by using ``renku workflow export``:

.. code-block:: console

    $ renku workflow export --format cwl my-run
    baseCommand:
    - cp
    class: CommandLineTool
    cwlVersion: v1.0
    id: 63e3a2a8-5b40-49b2-a2f4-eecc37bc76b0
    inputs:
    - default: B
    id: _plans_9a0961844fcc46e1816fde00f57e24a8_outputs_2_arg
    inputBinding:
        position: 2
    type: string
    - default:
        class: File
        location: file:///home/user/my-project/A
    id: _plans_9a0961844fcc46e1816fde00f57e24a8_inputs_1
    inputBinding:
        position: 1
    type: File
    - default:
        class: Directory
        location: file:///home/user/my-project/.renku
    id: input_renku_metadata
    type: Directory
    - default:
        class: Directory
        location: file:///home/user/my-project/.git
    id: input_git_directory
    type: Directory
    outputs:
    - id: _plans_9a0961844fcc46e1816fde00f57e24a8_outputs_2
    outputBinding:
        glob: $(inputs._plans_9a0961844fcc46e1816fde00f57e24a8_outputs_2_arg)
    type: File
    requirements:
    InitialWorkDirRequirement:
        listing:
        - entry: $(inputs._plans_9a0961844fcc46e1816fde00f57e24a8_inputs_1)
        entryname: A
        writable: false
        - entry: $(inputs.input_renku_metadata)
        entryname: .renku
        writable: false
        - entry: $(inputs.input_git_directory)
        entryname: .git
        writable: false

You can export into a file directly with ``-o <path>``.

.. cheatsheet::
   :group: Workflows
   :command: $ renku workflow export --format <format> <plan>
   :description: Export a Plan in a given format (e.g. 'cwl').
   :target: rp


Composing Plans into larger workflows
*************************************
.. image:: ../../_static/asciicasts/compose_plan.delay.gif
   :width: 850
   :alt: Composing Plans

For more complex workflows consisting of several steps, you can use the
``renku workflow compose`` command. This creates a new workflow that has
sub-steps.

The basic usage is:

.. code-block:: console

   $ renku run --name step1 -- cp input intermediate
   $ renku run --name step2 -- cp intermediate output
   $ renku workflow compose my-composed-workflow step1 step2

This would create a new workflow called ``my-composed-workflow`` that
consists of ``step1`` and ``step2`` as steps. This new workflow is just
like any other workflow in renku in that it can be executed, exported
or composed with other workflows.

.. cheatsheet::
   :group: Workflows
   :command: $ renku workflow compose <composed-name> <plan> <plan>
   :description: Create a new Plan composed of child Plans.
   :target: rp

Workflows can also be composed based on past Runs and their
inputs/outputs, using the ``--from`` and ``--to`` parameters. This finds
chains of Runs from inputs to outputs and then adds them to the
composed plan, applying mappings (see below) where appropriate to make
sure the correct values for execution are used in the composite. This
also means that all the parameters in the used plans are exposed on the
composed plan directly.
In the example above, this would be:

.. code-block:: console

   $ renku workflow compose --from input --to output my-composed-workflow

You can expose parameters of child steps on the parent workflow using
``--map``/``-m``  arguments followed by a mapping expression. Mapping expressions
take the form of ``<name>=<expression>`` where ``name`` is the name of the
property to be created on the parent workflow and expression points to one
or more fields on the child steps that should be mapped to this property.
The expressions come in two flavors, absolute references using the names
of workflows and properties, and relative references specifying the
position within a workflow.

An absolute expression in the example above could be ``step1.my_dataset``
to refer to the input, output or argument named ``my_dataset`` on the step
``step1``. A relative expression could be ``@step2.@output1`` to refer
to the first output of the second step of the composed workflow.

Valid relative expressions are ``@input<n>``, ``@output<n>`` and ``@param<n>``
for the nth input, output or argument of a step, respectively. For referring
to steps inside a composed workflow, you can use ``@step<n>``. For referencing
a mapping on a composed workflow, you can use ``@mapping<n>``. Of course, the
names of the objects for all these cases also work.

The expressions can also be combined using ``,`` if a mapping should point
to more than one parameter of a child step.

You can mix absolute and relative reference in the same expression, as you see
fit.

A full example of this would be:

.. code-block:: console

   $ renku workflow compose --map input_file=step1.@input2 \
       --map output_file=@step1.my-output,@step2.step2s_output \
       my-composed-workflow step1 step2

This would create a mapping called ``input_file`` on the parent workflow that
points to the second input of ``step1`` and a mapping called ``output_file``
that points to both the output ``my-output`` on ``step1`` and
``step2s_output`` on ``step2``.

You can also set default values for mappings, which override the default values
of the parameters they're pointing to by using the ``--set``/``-s`` parameter, for
instance:

.. code-block:: console

   $ renku workflow compose --map input_file=step1.@input2 \
       --set input_file=data.csv
       my-composed-workflow step1 step2


This would lead to ``data.csv`` being used for the second input of
``step1`` when ``my-composed-workflow`` is executed (if it isn't overridden
at execution time).

You can add a description to the mappings to make them more human-readable
by using the ``--describe-param``/``-p`` parameter, as shown here:

.. code-block:: console

   $ renku workflow compose --map input_file=step1.@input2 \
       -p input_file="The dataset to process"
       my-composed-workflow step1 step2

You can also expose all inputs, outputs or parameters of child steps by
using ``--map-inputs``, ``--map-outputs`` or ``--map-params``, respectively.

On execution, renku will automatically detect links between steps, if an input
of one step uses the same path as an output of another step, and execute
them in the correct order. Since this depends on what values are passed
at runtime, you might want to enforce a certain order of steps by explicitly
mapping outputs to inputs.

You can do that using the ``--link <source>=<sink>`` parameters, e.g.
``--link step1.@output1=step2.@input1``. This gets recorded on the
workflow template and forces ``step2.@input1`` to always be set to the same
path as ``step1.@output1``, irrespective of which values are passed at
execution time.

This way, you can ensure that the steps in your workflow are always executed
in the correct order and that the dependencies between steps are modeled
correctly.

Renku can also add links for you automatically based on the default values
of inputs and outputs, where inputs/outputs that have the same path get
linked in the composed run. To do this, pass the ``--link-all`` flag.

.. warning:: Due to workflows having to be directed acyclic graphs, cycles
   in the dependencies are not allowed. E.g. step1 depending on step2
   depending on step1 is not allowed. Additionally, the flow of information
   has to be from outputs to inputs or parameters, so you cannot map an input
   to an output, only the other way around.

Values on inputs/outputs/parameters get set according to the following
order of precedence (lower precedence first):

- Default value on a input/output/parameter
- Default value on a mapping to the input/output/parameter
- Value passed to a mapping to the input/output/parameter
- Value passed to the input/output/parameter
- Value propagated to an input from the source of a workflow link

Editing Plans
*************

.. image:: ../../_static/asciicasts/edit_plan.delay.gif
   :width: 850
   :alt: Editing Plans

Plans can be edited in some limited fashion, but we do not allow structural
changes, as that might cause issues with the reproducibility and provenance of
the project. If you want to do structural changes (e.g. adding/removing
parameters), we recommend you record a new plan instead.

You can change the name and description of Plans and of their parameters, as
well as changing default values of the parameters using the ``renku workflow
edit`` command:

.. code-block:: console

   $ renku workflow edit my-run --name new-run --description "my description"
     --rename-param input-1=my-input --set my-input=other-file.txt
     --describe-param my-input="My input parameter" my-run

This would rename the Plan ``my-run`` to ``new-run``, change its description,
rename its parameter ``input-1`` to ``my-input`` and set the default of this
parameter to ``other-file.txt`` and set its description.

+-----------------------+------------------------------------------------------+
| Option                | Description                                          |
+=======================+======================================================+
| ``-n, --name``        | Plan's name                                          |
+-----------------------+------------------------------------------------------+
| ``-d, --description`` | Plan's description.                                  |
+-----------------------+------------------------------------------------------+
| ``-s, --set``         | Set default value for a parameter.                   |
|                       | Accepted format is '<name>=<value>'                  |
+-----------------------+------------------------------------------------------+
| ``-m, --map``         | Add a new mapping on the Plan.                       |
|                       | Accepted format is '<name>=<name or expression>'     |
+-----------------------+------------------------------------------------------+
| ``-r, --rename-param``| Rename a parameter.                                  |
|                       | Accepted format is '<name>="new name"'               |
+-----------------------+------------------------------------------------------+
| ``-d,``               | Add a description for a parameter.                   |
| ``--describe-param``  | Accepted format is '<name>="description"'            |
+-----------------------+------------------------------------------------------+
| ``-m, --metadata``    | Path to file containing custom JSON-LD metadata to   |
|                       | be added to the dataset.                             |
+-----------------------+------------------------------------------------------+


.. cheatsheet::
   :group: Workflows
   :command: $ renku workflow edit <plan>
   :description: Create a new Plan composed of child Plans.
   :target: rp

Removing Plans
**************

Sometimes you might want to discard a recorded Plan or reuse its name with a
new Plan. In these cases, you can delete the old plan using ``renku workflow
remove <plan name>``. Once a Plan is removed, it doesn't show up in most renku
workflow commands.
``renku update`` ignores deleted Plans, but ``renku rerun`` will still rerun
them if needed, to ensure reproducibility.

.. cheatsheet::
   :group: Workflows
   :command: $ renku workflow delete <plan>
   :description: Remove a Plan.
   :target: rp

Working with Runs
~~~~~~~~~~~~~~~~~

Listing Runs
************

To get a view of what commands have been execute in the project, you can use
the ``renku log --workflows`` command:

.. code-block:: console

    $ renku log --workflows
    DATE                 TYPE  DESCRIPTION
    -------------------  ----  -------------
    2021-09-21 15:46:02  Run   cp A C
    2021-09-21 10:52:51  Run   cp A B

Refer to the documentation of the :ref:`cli-log` command for more details.

Visualizing Executions
**********************

.. image:: ../../_static/asciicasts/visualize_runs.delay.gif
   :width: 850
   :alt: Visualizing Runs

You can visualize past Runs made with renku using the ``renku workflow
visualize`` command.
This will show a directed graph of executions and how they are connected. This
way you can see exactly how a file was generated and what steps it involved.
It also supports an interactive mode that lets you explore the graph in a more
detailed way.

.. code-block:: console

   $ renku run echo "input" > input
   $ renku run cp input intermediate
   $ renku run cp intermediate output
   $ renku workflow visualize
        ╔════════════╗
        ║echo > input║
        ╚════════════╝
                *
                *
                *
            ┌─────┐
            │input│
            └─────┘
                *
                *
                *
    ╔═════════════════════╗
    ║cp input intermediate║
    ╚═════════════════════╝
                *
                *
                *
        ┌────────────┐
        │intermediate│
        └────────────┘
                *
                *
                *
    ╔══════════════════════╗
    ║cp intermediate output║
    ╚══════════════════════╝
                *
                *
                *
            ┌──────┐
            │output│
            └──────┘

    $ renku workflow visualize intermediate
        ╔════════════╗
        ║echo > input║
        ╚════════════╝
            *
            *
            *
            ┌─────┐
            │input│
            └─────┘
            *
            *
            *
    ╔═════════════════════╗
    ║cp input intermediate║
    ╚═════════════════════╝
            *
            *
            *
        ┌────────────┐
        │intermediate│
        └────────────┘
    $ renku workflow visualize --from intermediate
        ┌────────────┐
        │intermediate│
        └────────────┘
                *
                *
                *
    ╔══════════════════════╗
    ║cp intermediate output║
    ╚══════════════════════╝
                *
                *
                *
            ┌──────┐
            │output│
            └──────┘

You can also run in interactive mode using the ``--interactive`` flag.

.. code-block:: console

   $ renku workflow visualize --interactive

This will allow you to navigate between workflow execution and see details
by pressing the <Enter> key.

If you prefer to elaborate the output graph further, or if you wish to export
it for any reason, you can use the ``--format`` option to specify an output
format.

The following example generates the graph using the `dot` format. It can
be stored in a file or piped directly to any compatible tool. Here we
use the ``dot`` command line tool from graphviz to generate an SVG file.

.. code-block:: console

   $ renku workflow visualize --format dot <path> | dot -Tsvg > graph.svg

Use ``renku workflow visualize -h`` to see all available options.

.. cheatsheet::
   :group: Workflows
   :command: $ renku workflow visualize [--interactive]
   :description: Show linked workflows as a graph.
   :target: rp


Removing Runs
*************

Renku allows you to undo a Run in a project by using ``renku workflow revert
<activity ID>``. You can obtain <activity ID> from the ``renku log`` command.
If the deleted run generated some files, Renku either deletes these files (in
case there are no earlier versions of them and they are not used in other
activities) or revert them to their earlier versions. You can ask Renku to keep the
generated files and only delete the metadata by passing the ``--metadata-only``
option.

.. warning:: Renku only checks project's runs/plans to see if files are used.
   It doesn't check if files, that are going to be deleted, are added to a
   dataset for example. Make sure that the project doesn't use such files in
   other places or always use ``--metadata-only`` option when reverting a run.

If you want to delete a run along with its plan use the ``--plan`` option.
This only deletes the plan if it's not used by any other activity.

Renku won't remove a run if there are downstream runs that depend on it. The
reason is that removing a run will break the link between its upstream and
downstream runs. If this is not an issue for you or if you want to delete the
downstream runs later, then pass the ``--force`` option to make Renku delete
the run anyway.

.. cheatsheet::
   :group: Workflows
   :command: $ renku workflow revert <activity ID>
   :description: Undo a Run.
   :target: rp

.. cheatsheet::
   :group: Undo Renku Command
   :command: $ renku workflow revert <activity ID>
   :description: Undo a Renku workflow execution.
   :target: ui


Input and output files
~~~~~~~~~~~~~~~~~~~~~~

You can list input and output files generated in the repository by running
``renku workflow inputs`` and ``renku workflow outputs`` commands. Alternatively,
you can check if all paths specified as arguments are input or output files
respectively.

.. code-block:: console

   $ renku run wc < source.txt > result.wc
   $ renku workflow inputs
   source.txt
   $ renku workflow outputs
   result.wc
   $ renku workflow outputs source.txt
   $ echo $?  # last command finished with an error code
   1

.. cheatsheet::
   :group: Workflows
   :command: $ renku workflow inputs|||$ renku workflow outputs
   :description: Show input respectively output files used by workflows.
   :target: rp

"""

import json
import os
import shutil
import sys
from pathlib import Path

import click
from lazy_object_proxy import Proxy

import renku.ui.cli.utils.color as color
from renku.command.format.workflow import WORKFLOW_COLUMNS, WORKFLOW_FORMATS, WORKFLOW_VISUALIZE_FORMATS
from renku.command.util import ERROR
from renku.command.view_model.activity_graph import ACTIVITY_GRAPH_COLUMNS
from renku.core import errors
from renku.domain_model.constant import NO_VALUE
from renku.ui.cli.utils.callback import ClickCallback
from renku.ui.cli.utils.click import shell_complete_workflows
from renku.ui.cli.utils.plugins import available_workflow_providers, get_supported_formats
from renku.ui.cli.utils.terminal import print_workflow_file, show_text_with_pager


@click.group()
def workflow():
    """Workflow commands."""
    pass


@workflow.command("ls")
@click.option(
    "--format", type=click.Choice(list(WORKFLOW_FORMATS.keys())), default="tabular", help="Choose an output format."
)
@click.option(
    "-c",
    "--columns",
    type=click.STRING,
    default="id,name,command",
    metavar="<columns>",
    help="Comma-separated list of column to display: {}.".format(", ".join(WORKFLOW_COLUMNS.keys())),
    show_default=True,
)
def list_workflows(format, columns):
    """List or manage workflows with subcommands."""
    from renku.command.workflow import list_workflows_command

    result = list_workflows_command().build().execute(format=format, columns=columns)
    click.echo(result.output)


@workflow.command()
@click.argument("name_or_id_or_path", metavar="<name_or_id_or_path>", shell_complete=shell_complete_workflows)
def show(name_or_id_or_path):
    """Show details for workflow <name_or_id_or_path>."""
    from renku.command.view_model.plan import PlanViewModel
    from renku.command.view_model.workflow_file import WorkflowFileViewModel
    from renku.command.workflow import show_workflow_command
    from renku.ui.cli.utils.terminal import print_composite_plan, print_plan

    communicator = ClickCallback()

    plan = (
        show_workflow_command()
        .with_communicator(communicator)
        .build()
        .execute(name_or_id_or_path=name_or_id_or_path)
        .output
    )

    if plan:
        if isinstance(plan, WorkflowFileViewModel):
            print_workflow_file(plan)
        elif isinstance(plan, PlanViewModel):
            print_plan(plan)
        else:
            print_composite_plan(plan)
    else:
        click.secho(ERROR + f"Workflow '{name_or_id_or_path}' not found.")


@workflow.command()
@click.argument("name", metavar="<name>", shell_complete=shell_complete_workflows)
@click.option("--force", is_flag=True, help="Override the existence check.")
def remove(name, force):
    """Remove a workflow named <name>."""
    from renku.command.workflow import remove_plan_command

    communicator = ClickCallback()

    remove_plan_command().with_communicator(communicator).build().execute(name_or_id=name, force=force)


@workflow.command()
@click.option("-d", "--description", help="Workflow step's description.")
@click.option("mappings", "-m", "--map", multiple=True, help="Mapping for a workflow parameter.")
@click.option("defaults", "-s", "--set", multiple=True, help="Default value for a workflow parameter.")
@click.option("links", "-l", "--link", multiple=True, help="Link source and sink parameters to connect steps.")
@click.option("-p", "--describe-param", multiple=True, help="Add description for a workflow parameter.")
@click.option("--map-inputs", is_flag=True, help="Exposes all child inputs as inputs on the CompositePlan.")
@click.option("--map-outputs", is_flag=True, help="Exposes all child outputs as outputs on the CompositePlan.")
@click.option("--map-params", is_flag=True, help="Exposes all child parameters as parameters on the CompositePlan.")
@click.option("--map-all", is_flag=True, help="Combination of --map-inputs, --map-outputs, --map-params.")
@click.option("--link-all", is_flag=True, help="Automatically link steps based on default values.")
@click.option("--keyword", multiple=True, help="List of keywords for the workflow.")
@click.option(
    "--from",
    "sources",
    type=click.Path(exists=True, dir_okay=False),
    multiple=True,
    help="Start a composite plan from this file as input.",
)
@click.option(
    "--to",
    "sinks",
    type=click.Path(exists=True, dir_okay=True),
    multiple=True,
    help="End a composite plan at this file as output.",
)
@click.option(
    "--creator",
    "creators",
    default=None,
    multiple=True,
    type=click.UNPROCESSED,
    help="Creator's name, email, and affiliation. Accepted format is 'Forename Surname <email> [affiliation]'.",
)
@click.argument("name", required=True)
@click.argument("steps", nargs=-1, type=click.UNPROCESSED, shell_complete=shell_complete_workflows)
def compose(
    description,
    mappings,
    defaults,
    links,
    describe_param,
    map_inputs,
    map_outputs,
    map_params,
    map_all,
    link_all,
    keyword,
    sources,
    sinks,
    creators,
    name,
    steps,
):
    """Create a composite workflow consisting of multiple steps."""
    from renku.command.workflow import compose_workflow_command
    from renku.core.util.metadata import construct_creators
    from renku.ui.cli.utils.terminal import print_composite_plan

    if (sources or sinks) and steps:
        click.secho(ERROR + "--from/--to cannot be used at the same time as passing run/step names.")
        exit(1)
    elif not (sources or sinks or steps):
        click.secho(ERROR + "Either --from/--to passing run/step names is required.")
        exit(1)

    if creators:
        creators, _ = construct_creators(creators)

    if map_all:
        map_inputs = map_outputs = map_params = True

    result = (
        compose_workflow_command()
        .build()
        .execute(
            name=name,
            description=description,
            mappings=mappings,
            defaults=defaults,
            links=links,
            param_descriptions=describe_param,
            map_inputs=map_inputs,
            map_outputs=map_outputs,
            map_params=map_params,
            link_all=link_all,
            keywords=keyword,
            steps=steps,
            sources=sources,
            sinks=sinks,
            creators=creators,
        )
    )

    if not result.error:
        print_composite_plan(result.output)


@workflow.command()
@click.argument("workflow_name", metavar="<name or uuid>", shell_complete=shell_complete_workflows)
@click.option("-n", "--name", metavar="<new name>", help="New name of the workflow")
@click.option("-d", "--description", metavar="<new desc>", help="New description of the workflow")
@click.option(
    "-s",
    "--set",
    "set_params",
    multiple=True,
    metavar="<parameter>=<value>",
    help="Set default <value> for a <parameter>/add new parameter",
)
@click.option(
    "-m",
    "--map",
    "map_params",
    multiple=True,
    metavar="<parameter>=<parameter or expression>",
    help="New mapping on the workflow",
)
@click.option(
    "-r",
    "--rename-param",
    "rename_params",
    multiple=True,
    metavar='<parameter>="name"',
    help="New name for parameter",
)
@click.option(
    "--describe-param",
    "describe_params",
    multiple=True,
    metavar='<parameter>="description"',
    help="New description of the workflow",
)
@click.option(
    "-m",
    "--metadata",
    default=None,
    type=click.Path(exists=True, dir_okay=False),
    help="Custom metadata to be associated with the workflow.",
)
@click.option(
    "--creator",
    "creators",
    default=[NO_VALUE],
    multiple=True,
    type=click.UNPROCESSED,
    help="Creator's name, email, and affiliation. Accepted format is 'Forename Surname <email> [affiliation]'.",
)
@click.option(
    "--keyword",
    "keywords",
    default=[NO_VALUE],
    type=click.UNPROCESSED,
    multiple=True,
    help="List of keywords for the workflow.",
)
def edit(
    workflow_name,
    name,
    description,
    set_params,
    map_params,
    rename_params,
    describe_params,
    metadata,
    creators,
    keywords,
):
    """Edit workflow details."""
    from renku.command.view_model.plan import PlanViewModel
    from renku.command.workflow import edit_workflow_command
    from renku.core.util.metadata import construct_creators
    from renku.ui.cli.utils.terminal import print_composite_plan, print_plan

    if list(creators) == [NO_VALUE]:
        creators = NO_VALUE

    keywords = list(keywords)
    if keywords == [NO_VALUE]:
        keywords = NO_VALUE

    if creators and creators is not NO_VALUE:
        creators, _ = construct_creators(creators, ignore_email=True)

    custom_metadata = None

    if metadata:
        custom_metadata = json.loads(Path(metadata).read_text())

    result = (
        edit_workflow_command()
        .build()
        .execute(
            name=workflow_name,
            new_name=name,
            description=description,
            set_params=set_params,
            map_params=map_params,
            rename_params=rename_params,
            describe_params=describe_params,
            creators=creators,
            keywords=keywords,
            custom_metadata=custom_metadata,
        )
    )
    if not result.error:
        plan = result.output
        if isinstance(plan, PlanViewModel):
            print_plan(plan)
        else:
            print_composite_plan(plan)


@workflow.command()
@click.argument("workflow_name", metavar="<name or uuid>", shell_complete=shell_complete_workflows)
@click.option(
    "-f",
    "--format",
    default="cwl",
    type=click.Choice(Proxy(get_supported_formats), case_sensitive=False),
    show_default=True,
    help="Workflow language format.",
)
@click.option(
    "-o",
    "--output",
    metavar="<path>",
    type=click.Path(exists=False),
    default=None,
    help="Save to <path> instead of printing to terminal",
)
@click.option(
    "--values",
    metavar="<file>",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="YAML file containing parameter mappings to be used.",
)
def export(workflow_name, format, output, values):
    """Export workflow."""
    from renku.command.workflow import export_workflow_command
    from renku.core.util.os import safe_read_yaml

    communicator = ClickCallback()

    values_dict = None
    if values is not None:
        values_dict = safe_read_yaml(values)

    result = (
        export_workflow_command()
        .with_communicator(communicator)
        .build()
        .execute(
            name_or_id=workflow_name,
            format=format,
            output=output,
            values=values_dict,
            basedir=None,
            resolve_paths=True,
            nest_workflows=False,
        )
    )

    if not output:
        click.echo(result.output)


@workflow.command()
@click.argument("paths", type=click.Path(exists=True, dir_okay=True), nargs=-1)
@click.pass_context
def inputs(ctx, paths):
    """Show all inputs used by workflows.

    <PATHS>    Limit results to these paths.
    """
    from renku.command.workflow import workflow_inputs_command

    result = workflow_inputs_command().build().execute(paths=paths)

    input_paths = result.output

    click.echo("\n".join(input_paths))

    if paths:
        if not input_paths or any(
            p not in input_paths and all(Path(o) not in Path(p).parents for o in input_paths) for p in paths
        ):
            ctx.exit(1)


@workflow.command()
@click.argument("paths", type=click.Path(exists=True, dir_okay=True), nargs=-1)
@click.pass_context
def outputs(ctx, paths):
    """Show all outputs generated by workflows.

    <PATHS>    Limit results to these paths.
    """
    from renku.command.workflow import workflow_outputs_command

    result = workflow_outputs_command().build().execute(paths=paths)

    output_paths = result.output

    click.echo("\n".join(output_paths))

    if paths:
        if not output_paths or any(
            p not in output_paths and all(Path(o) not in Path(p).parents for o in output_paths) for p in paths
        ):
            ctx.exit(1)


@workflow.command()
@click.option(
    "provider",
    "-p",
    "--provider",
    default="toil",
    show_default=True,
    type=click.Choice(Proxy(available_workflow_providers), case_sensitive=False),
    help="The workflow engine to use.",
)
@click.option(
    "config", "-c", "--config", metavar="<config file>", help="YAML file containing configuration for the provider."
)
@click.option(
    "set_params",
    "-s",
    "--set",
    multiple=True,
    metavar="<parameter>=<value>",
    help="Set <value> for a <parameter> to be used in execution.",
)
@click.option(
    "--values",
    metavar="<values-file>",
    type=click.Path(exists=True, dir_okay=False),
    help="YAML file containing parameter mappings to be used.",
)
@click.option("--skip-metadata-update", is_flag=True, help="Do not update the metadata store for the execution.")
@click.argument("name_or_id", required=True, shell_complete=shell_complete_workflows)
def execute(
    provider,
    config,
    set_params,
    values,
    skip_metadata_update,
    name_or_id,
):
    """Execute a given workflow."""
    from renku.command.workflow import execute_workflow_command

    communicator = ClickCallback()

    result = (
        execute_workflow_command(skip_metadata_update=skip_metadata_update)
        .with_communicator(communicator)
        .build()
        .execute(
            name_or_id=name_or_id,
            provider=provider,
            config=config,
            values=values,
            set_params=set_params,
        )
    )

    if result.output:
        click.echo(
            "Unchanged files:\n\n\t{}".format("\n\t".join(click.style(path, fg=color.YELLOW) for path in result.output))
        )


@workflow.command()
@click.option(
    "--from",
    "sources",
    type=click.Path(exists=False, dir_okay=False),
    multiple=True,
    help="Start drawing the graph from this file.",
)
@click.option(
    "-c",
    "--columns",
    type=click.STRING,
    default="command",
    metavar="<columns>",
    help="Comma-separated list of column to display: {}.".format(", ".join(ACTIVITY_GRAPH_COLUMNS.keys())),
    show_default=True,
)
@click.option("-x", "--exclude-files", is_flag=True, help="Hide file nodes, only show Runs.")
@click.option("-a", "--ascii", is_flag=True, help="Only use Ascii characters for formatting.")
@click.option(
    "--revision",
    type=click.STRING,
    help="Git revision to generate the graph for.",
)
@click.option(
    "--format",
    type=click.Choice(list(WORKFLOW_VISUALIZE_FORMATS.keys())),
    default="console",
    help="Choose an output format.",
)
@click.option(
    "-i", "--interactive", is_flag=True, help="Interactively explore run graph. Only available for console output"
)
@click.option("--no-color", is_flag=True, help="Don't colorize console output.")
@click.option("--pager", is_flag=True, help="Force use pager (less) for console output.")
@click.option("--no-pager", is_flag=True, help="Don't use pager (less) for console output.")
@click.argument("paths", type=click.Path(exists=False, dir_okay=True), nargs=-1)
def visualize(sources, columns, exclude_files, ascii, revision, format, interactive, no_color, pager, no_pager, paths):
    """Visualization of workflows that produced outputs at the specified paths.

    Either PATHS or --from need to be set.
    """
    from renku.command.workflow import visualize_graph_command

    if format == WORKFLOW_VISUALIZE_FORMATS["console"] and pager and no_pager:
        raise errors.ParameterError("Can't use both --pager and --no-pager.")
    if revision and not paths:
        raise errors.ParameterError("Can't use --revision without specifying PATHS.")

    result = (
        visualize_graph_command()
        .build()
        .execute(sources=sources, targets=paths, show_files=not exclude_files, revision=revision)
    )
    if format == WORKFLOW_VISUALIZE_FORMATS["dot"]:
        output = result.output.dot_representation(columns=columns)

        if not output:
            return

        click.echo(output)
        return
    else:
        text_output, navigation_data = result.output.text_representation(
            columns=columns, color=not no_color, ascii=ascii
        )

        if not text_output:
            return

        if not interactive:
            max_width = max(node[1].x for layer in navigation_data for node in layer)
            tty_size = shutil.get_terminal_size(fallback=(120, 120))

            if no_pager or not sys.stdout.isatty() or os.system(f"less 2>{os.devnull}") != 0:  # nosec
                use_pager = False
            elif pager:
                use_pager = True
            elif max_width < tty_size.columns:
                use_pager = False
            else:
                use_pager = True

            if use_pager:
                show_text_with_pager(text_output)
            else:
                click.echo(text_output)
            return

        from renku.ui.cli.utils.curses import CursesActivityGraphViewer

        viewer = CursesActivityGraphViewer(
            text_output, navigation_data, result.output.vertical_space, use_color=not no_color
        )
        viewer.run()


@workflow.command()
@click.option("--skip-metadata-update", is_flag=True, help="Do not update the metadata store for the execution.")
@click.option(
    "mapping_path",
    "--mapping",
    metavar="<file>",
    type=click.Path(exists=True, dir_okay=False),
    help="YAML file containing parameter mappings to be used.",
)
@click.option(
    "--dry-run",
    "-n",
    is_flag=True,
    default=False,
    help="Print the generated plans with their parameters instead of executing.",
    show_default=True,
)
@click.option(
    "provider",
    "-p",
    "--provider",
    default="toil",
    show_default=True,
    type=click.Choice(Proxy(available_workflow_providers), case_sensitive=False),
    help="The workflow engine to use.",
)
@click.option("mappings", "-m", "--map", multiple=True, help="Mapping for a workflow parameter.")
@click.option("config", "-c", "--config", metavar="<config file>", help="YAML file containing config for the provider.")
@click.argument("name_or_id", required=True, shell_complete=shell_complete_workflows)
def iterate(name_or_id, mappings, mapping_path, dry_run, provider, config, skip_metadata_update):
    """Execute a workflow by iterating through a range of provided parameters."""
    from renku.command.view_model.plan import PlanViewModel
    from renku.command.workflow import iterate_workflow_command, show_workflow_command
    from renku.ui.cli.utils.terminal import print_composite_plan, print_plan

    if len(mappings) == 0 and mapping_path is None:
        raise errors.UsageError("No mapping has been given for the iteration!")

    plan = show_workflow_command().build().execute(name_or_id_or_path=name_or_id).output

    if plan:
        if isinstance(plan, PlanViewModel):
            print_plan(plan)
        else:
            print_composite_plan(plan)

    communicator = ClickCallback()
    iterate_workflow_command(skip_metadata_update=skip_metadata_update).with_communicator(communicator).build().execute(
        name_or_id=name_or_id,
        mapping_path=mapping_path,
        mappings=mappings,
        dry_run=dry_run,
        provider=provider,
        config=config,
    )


@workflow.command()
@click.option(
    "-m",
    "--metadata-only",
    default=False,
    help="Only undo metadata, leave generated outputs unchanged.",
    is_flag=True,
    show_default=True,
)
@click.option(
    "-f",
    "--force",
    default=False,
    help="Force-revert the activity, even if it breaks things.",
    is_flag=True,
    show_default=True,
)
@click.option(
    "-p",
    "--plan",
    default=False,
    help="Delete activity's plan if no other activity is using it.",
    is_flag=True,
    show_default=True,
)
@click.argument("activity_id", required=True)
def revert(metadata_only, force, plan, activity_id):
    """Revert activity metadata and generations."""
    from renku.command.workflow import revert_activity_command

    communicator = ClickCallback()
    try:
        revert_activity_command().with_communicator(communicator).build().execute(
            metadata_only=metadata_only, force=force, delete_plan=plan, activity_id=activity_id
        )
    except errors.ActivityDownstreamNotEmptyError:
        raise errors.ParameterError(
            "Activity has downstream dependent activities: Pass '--force' if you want to revert the activity anyways."
        )



#
# Copyright 2017-2023 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku plan database gateway implementation."""

from typing import Dict, List, Optional, cast

import deal

from renku.core import errors
from renku.core.interface.plan_gateway import IPlanGateway
from renku.domain_model.project_context import project_context
from renku.domain_model.workflow.plan import AbstractPlan


class PlanGateway(IPlanGateway):
    """Gateway for plan database operations."""

    def get_by_id(self, id: Optional[str]) -> Optional[AbstractPlan]:
        """Get a plan by id."""
        return project_context.database["plans"].get(id)

    def get_by_name(self, name: str) -> Optional[AbstractPlan]:
        """Get a plan by name."""
        return project_context.database["plans-by-name"].get(name)

    def get_by_name_or_id(self, name_or_id: str) -> AbstractPlan:
        """Get a plan by name or id."""
        workflow = self.get_by_id(name_or_id) or self.get_by_name(name_or_id)

        if not workflow:
            raise errors.WorkflowNotFoundError(name_or_id)
        return workflow

    def list_by_name(self, starts_with: str, ends_with: Optional[str] = None) -> List[str]:
        """Search plans by name."""
        return [
            name
            for name in project_context.database["plans-by-name"].keys(min=starts_with, max=ends_with)
            if not cast(AbstractPlan, self.get_by_name(name)).deleted
        ]

    def get_newest_plans_by_names(self, include_deleted: bool = False) -> Dict[str, AbstractPlan]:
        """Return a mapping of all plan names to their newest plans."""
        database = project_context.database
        if include_deleted:
            return dict(database["plans-by-name"])
        return {k: v for k, v in database["plans-by-name"].items() if not v.deleted}

    def get_all_plans(self) -> List[AbstractPlan]:
        """Get all plans in project."""
        return list(project_context.database["plans"].values())

    @deal.pre(lambda _: _.plan.date_created is not None)
    @deal.pre(lambda _: _.plan.date_created >= project_context.project.date_created)
    @deal.pre(lambda _: _.plan.date_modified is None or _.plan.date_modified >= project_context.project.date_created)
    @deal.pre(lambda _: _.plan.date_removed is None or _.plan.date_removed >= project_context.project.date_created)
    def add(self, plan: AbstractPlan) -> None:
        """Add a plan to the database."""
        database = project_context.database

        if database["plans"].get(plan.id) is not None:
            return

        database["plans"].add(plan)

        if plan.derived_from is not None:
            derived_from = self.get_by_id(plan.derived_from)

            if derived_from is not None:
                database["plans-by-name"].pop(derived_from.name, None)
        database["plans-by-name"].add(plan)
